# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_bincount(backends):
    inputs = [1, 2, 3, 3, 5]
    for B in backends:
        t = B.tensor(inputs)
        cnt = B.bincount(t, minlength=6)
        assert cnt.shape[0] == 6
        assert B.tensor_equal(cnt, B.tensor([0, 1, 1, 2, 0, 1]))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def box(a):
    "box single tensor into a list"
    if isinstance(a, list):
        return a
    else:
        return [a]


def unbox(a):
    "Unbox list with a single element"
    if isinstance(a, list) and len(a) == 1:
        return a[0]
    else:
        return a



import json
import tensorflow as tf
from collections import defaultdict
import numpy as np

import evoflow.backend as B
from evoflow.engine import EvoFlow
from evoflow.engine import FitnessFunction
from evoflow.selection import SelectFittest
from evoflow.population import uniform_population
from evoflow.ops import Input, Shuffle, Reverse1D


class TSPFitness(FitnessFunction):
    def __init__(self,
                 distances,
                 num_cities,
                 baseline_distance=0,
                 penality=100000,
                 **kwargs):
        """
        """
        self.num_cities = num_cities
        self.distances = B.flatten(distances)
        self.penality = penality
        self.baseline_distance = int(baseline_distance)
        super(TSPFitness, self).__init__(**kwargs)

    def call(self, population, normalize=True):
        """
            Parallel lookup and distance computation:
            - multiply the tensor by population shifed by 1 which gives
            the id to lookup in the flat
            distance array
            - reduce_sum for the total distance
            - 1/reduce_sum so fitness goes from 0 to 1
        """

        shifted_population = B.roll(population, 1, axis=1)
        idxs = (population * self.num_cities) + shifted_population
        distances = B.take(self.distances, idxs)

        # total distance
        total_distance = B.sum(distances, axis=1)
        return total_distance


def tsp_setup(num_cities):

    # get files
    zip_fname = "tsp_%s.zip" % num_cities
    origin = "https://storage.googleapis.com/evoflow/datasets/tsp/cities_%s.zip" % num_cities  # noqa
    download_path = tf.keras.utils.get_file(zip_fname, origin, extract=True)

    # process city info
    json_fname = "%s/cities_%s.json" % (download_path.replace(zip_fname,
                                                              ''), num_cities)
    cities = json.loads(open(json_fname).read())

    idx2city = {}
    for city in cities:
        idx2city[city['idx']] = city

    chart_data = defaultdict(list)
    for city in cities:
        chart_data['lat'].append(city['lat'])
        chart_data['lon'].append(city['lon'])
        chart_data['name'].append(city['name'])
        chart_data['population'].append(city['population'])

    distance_fname = "%sdistances_%s.npz" % (download_path.replace(
        zip_fname, ''), num_cities)
    distances = np.load(distance_fname)['distances']
    distances = distances.astype(B.intx())
    return cities, chart_data, distances, idx2city


def test_solve_tsp():
    population_shape = (100, 10)
    generations = 30

    cities, chart_data, distances, idx2city = tsp_setup(10)
    BASELINE = 6000
    NUM_CITIES = len(cities)
    NUM_REVERSE_OPERATIONS = 4
    MAX_REVERSE_PROBABILITY = 0.3
    REVERSE_POPULATION_FRACTION = 0.3
    MIN_REVERSE_PROBABILITY = 0.1

    SHUFFLE_POPULATION_FRACTION = 0.2

    population = uniform_population(population_shape)
    rpi = MAX_REVERSE_PROBABILITY / NUM_REVERSE_OPERATIONS
    reverse_probabilty = 1 - rpi

    # Evolution model

    inputs = Input(shape=population.shape)
    x = inputs
    for idx in range(NUM_REVERSE_OPERATIONS):
        x = Reverse1D(population_fraction=REVERSE_POPULATION_FRACTION,
                      max_reverse_probability=reverse_probabilty)(x)
        reverse_probabilty = max(reverse_probabilty - rpi,
                                 MIN_REVERSE_PROBABILITY)

    x = Shuffle(population_fraction=SHUFFLE_POPULATION_FRACTION)(x)
    outputs = x
    ef = EvoFlow(inputs, outputs, debug=False)

    evolution_strategy = SelectFittest(mode='min')
    fitness_fn = TSPFitness(distances, NUM_CITIES)
    ef.compile(evolution_strategy, fitness_fn)
    results = ef.evolve(population, generations=generations, verbose=0)
    metrics = results.get_latest_metrics()
    assert metrics['Fitness function']['min'] < BASELINE



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import evoflow.backend as B
from evoflow.ops import Dummy, Input


def test_eager():
    "when passing concrete value GF is suppposed to return a concrete value"
    val = B.tensor(42)
    assert Dummy(debug=True)(val) == val


def test_graph():
    "When passing OPs GF is supposed to return a graph"
    i = Input((42, 1))
    d1 = Dummy()(i)
    d2 = Dummy()(i)
    r = Dummy()([d1, d2])
    assert issubclass(type(r), Dummy)
    assert r.call(42) == 42  # explicitly execute the graph op



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from evoflow.config import set_backend, get_backend

# override existing setting to allow reload
if 'GENEFLOW_BACKEND' in os.environ:
    set_backend(os.environ['GENEFLOW_BACKEND'])
elif not get_backend():
    # check if we find cupy
    try:
        import tensorflow  # noqa: F403, F401
    except ImportError:
        set_backend('numpy')
    else:
        set_backend('tensorflow')

if get_backend() == 'numpy':
    from .numpy import *  # noqa: F403, F401
elif get_backend() == 'tensorflow':
    from .tensorflow import *  # noqa: F403, F401
    # ensure we don't run out of memory
    import tensorflow as tf
    gpus = tf.config.experimental.list_physical_devices('GPU')
    if gpus:
        for gpu in gpus:
            try:
                tf.config.experimental.set_memory_growth(gpu, True)
            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)

            logical_gpus = tf.config.experimental.list_logical_devices('GPU')
            msg = "GPUs Physical: %d - Logical GPU:%d\n" % (len(gpus),
                                                            len(logical_gpus))
            sys.stderr.write(msg)

elif get_backend() == 'cupy':
    from .cupy import *  # noqa: F403, F401

sys.stderr.write('Using %s backend\n' % get_backend())



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# base
from ..engine.op import OP  # noqa: F401

# core
from .core import Dummy  # noqa: F401

# inputs
from ..engine.input import Input  # noqa: F401

# Shuffle
from .shuffle import Shuffle  # noqa: F401

# Reverse
from .reverse import Reverse1D  # noqa: F401
from .reverse import Reverse2D  # noqa: F401
from .reverse import Reverse3D  # noqa: F401

# mutation
from .random_mutation import RandomMutations1D  # noqa: F401
from .random_mutation import RandomMutations2D  # noqa: F401
from .random_mutation import RandomMutations3D  # noqa: F401

# crossover
from .uniform_crossover import UniformCrossover1D  # noqa: F401
from .uniform_crossover import UniformCrossover2D  # noqa: F401
from .uniform_crossover import UniformCrossover3D  # noqa: F401
from .single_crossover import SingleCrossover1D  # noqa: F401
from .single_crossover import SingleCrossover2D  # noqa: F401
from .single_crossover import SingleCrossover3D  # noqa: F401
from .dual_crossover import DualCrossover1D  # noqa: F401
from .dual_crossover import DualCrossover2D  # noqa:F401
from .dual_crossover import DualCrossover3D  # noqa: F401



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from termcolor import cprint
# logging
import logging
logger = logging.getLogger()

# need to be done as early as possible
try:
    import tensorflow as tf
except:  # noqa
    cprint('Tensoflow not found', 'yellow')
    pass
else:
    tf.config.set_visible_devices([], 'GPU')
    cprint('Tensoflow set to CPU', 'green')


def pytest_configure(config):
    backend = config.option.backend
    if backend:
        from evoflow.config import set_backend
        set_backend(backend)
        cprint('Requested backend: %s' % backend, 'magenta')
    else:
        cprint("Requested backend: default (tensorflow)", 'magenta')


def pytest_addoption(parser):
    parser.addoption("--backend",
                     help="specify the backend: numpy, cupy, tensorflow")


@pytest.fixture(scope="session")
def backends():
    """constructing a list of backend

    # ! only use this for backend testing. Higher level function import the
    # ! backend so it will create type conflict.
    """
    import evoflow.backend.numpy as NP

    try:
        import cupy as cp  # noqa
        logger.info('cupy found')
    except:  # noqa
        import evoflow.backend.numpy as CP
        logger.info('cupy not found - using numpy instead')
    else:
        import evoflow.backend.cupy as CP

    try:
        import tensorflow  # noqa
        logger.info('TensorFlow found')
    except:  # noqa
        import evoflow.backend.numpy as TF
        logger.info('TensorFlow not found - using numpy instead')
    else:
        import evoflow.backend.tensorflow as TF

    return [NP, CP, TF]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.fitness import InvertedCosineSimilarity
import evoflow.backend as B


def test_cosine_similarity_1dpopulation(backends):
    "test the function works on a population and broadcast correctly"

    reference = [1, 0, 1, 1, 0, 1, 1, 1]
    different = [1, 1, 1, 0, 1, 1, 1, 1]

    r = B.tensor(reference)
    d = B.tensor(different)
    population = B.tensor([r, d, r])
    cs = InvertedCosineSimilarity(r)

    # similar vector have a distance of 1
    distances = cs(population)
    print("distances", distances)
    assert B.assert_near(distances[0], 1, absolute_tolerance=0.001)
    assert not B.assert_near(distances[1], 1, absolute_tolerance=0.001)
    assert B.assert_near(distances[2], 1, absolute_tolerance=0.001)


def test_cosine_2d(backends):
    INSERTION_POINTS = [0, 10, 20]  # where we copy the ref chromosome

    ref_chromosome = B.randint(0, 2, (32, 32))
    ref_pop = B.randint(0, 2, (64, 32, 32))

    ref_pop = B.as_numpy_array(ref_pop)
    inserstion = B.as_numpy_array(ref_chromosome)
    for idx in INSERTION_POINTS:
        ref_pop[idx] = inserstion

    ref_pop = B.tensor(ref_pop)

    cs = InvertedCosineSimilarity(ref_chromosome)
    distances = cs(ref_pop)
    print(distances)

    for idx, dst in enumerate(distances):
        if idx in INSERTION_POINTS:
            assert B.assert_near(dst, 1.0, absolute_tolerance=0.001)
        else:
            assert dst < 1
            assert dst > 0



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import time
from collections import defaultdict
import altair as alt
from tabulate import tabulate
import pandas as pd
import evoflow.backend as B
from evoflow.utils import unbox
from evoflow.io import print_debug


class Results(object):
    def __init__(self, debug=False):

        self.start_time = time()

        self._populations = None
        self._fitness_scores = None
        self._metrics_latest = defaultdict(dict)  # convinience holder
        self._metrics_history = defaultdict(lambda: defaultdict(list))
        self.debug = debug

    def get_populations(self):
        return unbox(self._populations)

    def set_population(self, populations):
        """Assign unboxed population

        Args:
        populations (list): list of population tensors

        Note: the whole class assumes populations as a list so don't set
        unboxed results or it will break everything.
        """

        if self.debug:
            print_debug(populations)

        self._populations = populations

    def display_populations(self,
                            top_k=10,
                            max_chromosome_len=20,
                            precision=None):
        """Display the population

        Args:
            top_k (int, optional): Number of chromosomes to display.
            Defaults to 10.

            expected_max_value (int, optional): how many gene to display per
            chromosomes.

            precision (int, optional): how many digits per chromosome to
            display. Default to None. If None full value.

        FIXME: use HTML while in notebook
        """

        for pop_idx, population in enumerate(self._populations):
            rows = []
            for cidx, chromosome in enumerate(population):
                genes = []
                for gene in chromosome[:max_chromosome_len]:
                    if isinstance(precision, type(None)):
                        genes.append(str(gene))
                    elif precision > 0:
                        genes.append(str(round(float(gene), precision)))
                    else:
                        genes.append(str(int(gene)))

                if len(genes) != len(chromosome):
                    genes.append(' ...')

                genes = " ".join(genes)
                row = [self._fitness_scores[pop_idx][cidx], genes]
                if cidx == top_k:
                    break
                rows.append(row)
            print(
                tabulate(
                    rows,
                    headers=['fit score',
                             'genes [:%d]' % max_chromosome_len]))

    def plot_fitness(self):
        """Plots the various metrics"""
        return self._build_plot('Fitness function')

    def plot(self, group_name):
        """Plots group metrics"""
        return self._build_plot(group_name)

    def _build_plot(self, metric_group):
        "Build an altair plot for a given metric group"
        metrics = self.get_metrics_history()
        if metric_group not in metrics:
            raise ValueError("can't find metric group")

        data = metrics[metric_group]
        rows = []
        for name, values in data.items():
            for idx, val in enumerate(values):
                rows.append({'generation': idx, 'metric': name, 'value': val})
        metrics_pd = pd.DataFrame(rows)
        chart = alt.Chart(metrics_pd,
                          title='Fitness function').mark_line().encode(
                              x='generation', y='value',
                              color='metric').configure_title(fontSize=16,
                                                              offset=5,
                                                              orient='top',
                                                              anchor='middle')
        return chart

    def get_metrics_history(self):
        """Get the last evolution metrics values

        Returns:
            dict: name: value as list(float).
        """
        return self._metrics_history

    def get_latest_metrics(self, flatten=False):
        """Get the last evolution metrics values

        Args:
            flatten (bool, optional): Return metrics as a flat dictionary
            instead of a nested one

        Returns:
            dict: name:value as float.
        """

        if flatten:
            metrics = {}
            for group_name, group_data in self._metrics_latest.items():
                for metric, value in group_data.items():
                    k = "%s_%s" % (group_name, metric)
                    metrics[k] = value
            return metrics
        else:
            return self._metrics_latest

    def record_metrics(self, metrics):
        """Record metrics and track their history

        Args:
            metrics (dict(dict)): group of metrics to track. Of the form:
            [group][metric] = float(value)
        """
        for group, data in metrics.items():
            for metric, value in data.items():
                self._metrics_history[group][metric].append(value)
                self._metrics_latest[group][metric] = value

    def record_fitness(self, fitness_scores):
        """Compute and record fitness related metrics
        Args:
            fitness_scores (list(ndarray)): tensor holding fitness scores.
        """

        self._fitness_scores = fitness_scores

        # update history
        for pop_idx, fit_scores in enumerate(fitness_scores):
            METRICS = [['mean', B.mean], ['max', B.max], ['min', B.min]]

            for name, fn in METRICS:

                # only suffix is more than one population
                if len(fitness_scores) > 1:
                    name += "_%s" % pop_idx

                # compute result
                value = float(fn(fit_scores))

                self._metrics_history['Fitness function'][name].append(value)
                self._metrics_latest['Fitness function'][name] = value



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import evoflow.backend as B
from evoflow.engine import FitnessFunction


class InvertedCosineSimilarity(FitnessFunction):
    def __init__(self, reference_chromosome, **kwargs):
        """Inverted Cosine similarity function that returns 1 when chromosomes
        are similar to the reference chromose and [0, 1[ when different

        For reference implementation see
        https://github.com/scipy/scipy/blob/v0.14.0/scipy/spatial/distance.py#L267  # noqa

        Args:
            reference_chromosome (tensor1D): reference_chromosome.
        """
        super(InvertedCosineSimilarity, self).__init__(**kwargs)

        # cache ref chromosome flattend
        self.ref_chromosome = B.flatten(B.tensor(reference_chromosome))

        # caching ref pop norm
        self.ref_norm = B.norm(B.cast(self.ref_chromosome, B.floatx()))

    def call(self, population):
        # cosine only works on 1D array we need to flatten
        flat_pop = self._flatten_population(population)

        # numerator
        numerator = B.cast(B.dot(flat_pop, self.ref_chromosome), B.floatx())
        # denominator (norm(u) *  norm(v))
        # ! B.norm is not broadcasted we need our own version
        population_norm = B.broadcasted_norm(flat_pop)

        # population_norm = B.norm(B.cast(flat_pop, B.floatx()), axis=0)
        denominator = population_norm * self.ref_norm

        return (numerator / denominator)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_dot(backends):
    inputs = [1, 2, 3]
    inputs2 = [3, 2, 3]

    for B in backends:
        tensor = B.tensor(inputs)
        tensor2 = B.tensor(inputs2)
        product = B.dot(tensor, tensor2)
        assert product == 16


def test_dot_scalar(backends):
    inputs = [1, 2, 3]
    scalar = 2

    for B in backends:
        scalar_t = B.tensor(scalar)
        tensor = B.tensor(inputs)
        product = B.dot(tensor, scalar_t)
        assert B.tensor_equal(product, B.tensor([2, 4, 6]))


def test_dot_scalar_scalar(backends):
    scalar = 2
    scalar2 = 3

    for B in backends:
        t = B.tensor(scalar)
        t2 = B.tensor(scalar2)
        product = B.dot(t, t2)
        assert B.tensor_equal(product, B.tensor(6))


def test_dot_2d_2d(backends):
    arr = [[1, 2, 3], [3, 2, 1]]
    for B in backends:
        t = B.tensor(arr)
        t2 = B.transpose(t)
        product = B.dot(t, t2)
        assert B.tensor_equal(product, B.tensor([[14, 10], [10, 14]]))


def test_dot_2d_1d(backends):
    arr = [[1, 0, 1, 1, 0, 1, 1, 1], [1, 1, 1, 0, 1, 1, 1, 1],
           [1, 0, 1, 1, 0, 1, 1, 1]]
    arr2 = [1, 0, 1, 1, 0, 1, 1, 1]
    for B in backends:
        t = B.tensor(arr)
        t2 = B.tensor(arr2)
        product = B.dot(t, t2)
        print('product', product)
        expected = B.tensor([6, 5, 6])
        print('expected', expected)
        assert B.tensor_equal(product, expected)


def test_add(backends):
    inputs = [1, 2, 3]
    inputs2 = [1, 2, 3]
    expected = [2, 4, 6]
    for B in backends:
        tensor = B.tensor(inputs)
        tensor2 = B.tensor(inputs2)
        assert B.tensor_equal(B.add(tensor, tensor2), B.tensor(expected))


def test_substract(backends):
    inputs = [3, 6, 3]
    inputs2 = [1, 2, 6]
    expected = [2, 4, -3]
    for B in backends:
        tensor = B.tensor(inputs)
        tensor2 = B.tensor(inputs2)
        assert B.tensor_equal(B.subtract(tensor, tensor2), B.tensor(expected))


def test_multiply(backends):
    inputs = [3, 6, 3]
    inputs2 = [1, 2, -6]
    expected = [3, 12, -18]
    for B in backends:
        tensor = B.tensor(inputs)
        tensor2 = B.tensor(inputs2)
        assert B.tensor_equal(B.multiply(tensor, tensor2), B.tensor(expected))


def test_divide(backends):
    inputs = [3, 6, -3]
    inputs2 = [1, 2, 6]
    expected = [3, 3, -0.5]
    for B in backends:
        tensor = B.tensor(inputs)
        tensor2 = B.tensor(inputs2)
        div_results = B.divide(tensor, tensor2)
        print('result', div_results)
        expected_tensor = B.tensor(expected)
        print('expected', expected_tensor)
        assert B.tensor_equal(div_results, expected_tensor)


def test_mod(backends):
    inputs = [5, 6, -3]
    inputs2 = [3, 2, 6]
    expected = [2, 0, 3]
    for B in backends:
        tensor = B.tensor(inputs)
        tensor2 = B.tensor(inputs2)
        assert B.tensor_equal(B.mod(tensor, tensor2), B.tensor(expected))


def test_norm(backends):
    inputs = [1, 2, 3]
    expected = 3.7416573867739413
    for B in backends:
        tensor = B.tensor(inputs)
        norm = B.norm(B.cast(tensor, 'float32'))
        assert norm - expected < 0.0001



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx as nx
from termcolor import cprint
from tabulate import tabulate
from collections import defaultdict
from tqdm.auto import tqdm

from evoflow.utils import box
from evoflow.io import print_debug
from .results import Results

import tensorflow as tf


class EvoFlow(object):
    def __init__(self, inputs, outputs, debug=False):

        if not inputs:
            raise ValueError("Inputs can't be empty")
        if not outputs:
            raise ValueError("Ouputs can't be empty")

        # FIXME: check we have the same number of inputs and output
        # and they have the same shape because we are using the ouputs as
        # the next inputs

        # set debug
        self.debug = debug

        # graph underlying structure
        self.graph = nx.DiGraph()

        # tracking structures
        self.idx2op = {}  # op object
        self.idx2results = {}  # op computation result
        self.idx2input_idx = defaultdict(set)  # ops used as inputs
        self.idx2ouput_ops = defaultdict(set)  # ops used as outputs

        self.inputs_idx = []  # track which op idx are inputs
        self.outputs_idx = []  # track what op idx are outputs

        self.fitness = None
        self.compiled = False
        self._results = None

        # storing inputs tensors
        self.inputs = box(inputs)
        for ipt in self.inputs:
            self.inputs_idx.append(ipt.idx)

        # output
        self.outputs = box(outputs)
        for output in self.outputs:
            self.outputs_idx.append(output.idx)

        # build forward graph
        for output in self.outputs:
            self._add_op_to_graph(output, None, self.debug)

        # FIXME: check that the graph is fully connected from input to output

        # coerce exec_path as a list to allow reuse accros batches.
        self.execution_path = list(nx.topological_sort(self.graph))

    def compile(self, selection_strategy, fitness_functions):
        """Configure evoluationary model for training

        """
        # FIXME: check args validity
        self.selection_strategy = selection_strategy
        self.fitness_functions = box(fitness_functions)
        self.compiled = True

        self._results = Results(debug=self.debug)

    def history(self):
        return self._history

    def evolve(self, populations, generations=1, callbacks=None, verbose=1):

        if not self.compiled:
            raise ValueError("compile() must be run before using the graph")
            return

        populations = box(populations)
        self.print_debug("Initial Populations", populations)

        if not len(populations) == len(self.inputs):
            raise ValueError('The numbers of population must be equal\
                 to number of inputs')

        # assign initial value to inputs
        current_populations = []
        for pop_idx, ipt in enumerate(self.inputs):
            self.inputs[pop_idx].assign(populations[pop_idx])
            pop = self.inputs[pop_idx].get()
            current_populations.append(pop)
        num_populations = len(current_populations)
        self.print_debug('Initial current_populations', current_populations)

        # progress bar
        if verbose:
            pb = tqdm(total=generations, unit='generation')

        # evolve loop
        for generation_idx in range(generations):

            # callbacks
            if callbacks:
                for callback in callbacks:
                    callback.on_generation_begin(generation_idx)

            # perform evolution
            evolved_populations = self.perform_evolution()

            # assign evolved populations
            self.print_debug(generation_idx, 'evolved pop',
                             evolved_populations)

            fitness_scores_list = []  # keep track of fitness score accros pops
            for pop_idx in range(num_populations):
                # find current informaiton
                current_pop = current_populations[pop_idx]
                evolved_pop = evolved_populations[pop_idx]
                fitness_function = self.fitness_functions[pop_idx]

                self.print_debug('current_population', pop_idx, current_pop)
                self.print_debug('evolved_population', pop_idx, evolved_pop)
                # select population
                new_population, fitness_scores, metrics = self.selection_strategy(  # noqa
                    fitness_function, current_pop, evolved_pop)

                # update metrics
                self._results.record_metrics(metrics)

                # update population tensor
                self.inputs[pop_idx].assign(new_population)

                # track current population
                current_populations[pop_idx] = new_population
                # record fitness scores
                fitness_scores_list.append(fitness_scores)

            self._results.record_fitness(fitness_scores_list)

            latest_metrics = self._results.get_latest_metrics(flatten=True)

            # callbacks
            if callbacks:
                for callback in callbacks:
                    callback.on_generation_end(generation_idx, latest_metrics,
                                               fitness_scores_list,
                                               populations)

            # progress bar
            if verbose:
                formatted_metrics = {}
                for name, value in latest_metrics.items():
                    name = name.lower().replace(' ', '_')
                    formatted_metrics[name] = value
                pb.set_postfix(formatted_metrics)
                pb.update()

        if verbose:
            pb.close()

        # record last population
        self._results.set_population(current_populations)

        # return last evolution
        return self._results

    def perform_evolution(self):
        """
        Evolve population

        Args:
            populations (list): populations to evolve.
        """

        # single batch # FIXME: move to a batch function as we need for
        # evaluate
        self.print_debug('execution path:', self.execution_path)

        for op_idx in self.execution_path:
            op = self.idx2op[op_idx]

            self.print_debug('|- %s(%s)' %
                             (op.idx, self.idx2input_idx[op.idx]))

            # fetching inputs values
            inputs = []
            for input_idx in self.idx2input_idx[op_idx]:
                inputs.append(self.idx2results[input_idx])
            self.idx2results[op_idx] = op._call_from_graph(inputs)

        self.print_debug('idx2results:', self.idx2results.keys())

        # collect results
        results = []

        for op_idx in self.outputs_idx:
            results.append(self.idx2results[op_idx])

        return results

    def summary(self):
        "print a summary of the data flow"
        rows = []

        for op_idx in self.execution_path:
            op = self.idx2op[op_idx]

            # inputs
            if len(self.idx2input_idx[op_idx]):
                inputs = ",".join([o for o in self.idx2input_idx[op_idx]])
            else:
                inputs = ''

            # the op id and its shape
            op_info = "%s (%s)" % (op.idx, op.op_type)

            # shape
            op_shape = "%s" % str(op.get_output_shapes())

            # output
            # if len(self.idx2ouput_ops[op_idx]):
            #     outputs = ",".join([o for o in self.idx2ouput_ops[op_idx]])  # noqa E501
            # else:
            #     outputs = '>>'

            rows.append([op_info, op_shape, inputs])
        print(tabulate(rows, headers=['OP (type)', 'Output Shape', 'Inputs']))

    def print_debug(self, *msg):
        "output debug message"
        if self.debug:
            print_debug('GeneFlow', *msg)

    def _add_op_to_graph(self, op, output_op=None, debug=0):
        """
        Recursively insert nodes in the DAG

        Take the opportunity to compute outbound nodes and various variables

        Args:
            op (Op): the operation to insert
            output_op (Op): the output_op to add.
            debug (int): print debug
        """

        self.idx2op[op.idx] = op
        self.graph.add_node(op.idx)

        # recording inbound op if any
        if output_op:
            self.idx2ouput_ops[op.idx].add(output_op.idx)

        # Reversing the graph and storing it - notice:in_op become op
        for in_op in op.input_ops:

            self.idx2input_idx[op.idx].add(in_op.idx)

            if debug:
                cprint("[edge] %s->%s" % (in_op.idx, op.idx), 'yellow')

            self.graph.add_edge(in_op.idx, op.idx)
            self._add_op_to_graph(in_op, op, self.debug)  # up the chain



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from .common import _infer_dtype, intx, floatx


# -initialization-
def tensor(a, dtype=None):
    """Converts an object to a tensor.

    Args:
        a: The source object.
        dtype: Data type specifier. It is inferred from the input by default.
    Returns:
        ndarray: An array on the current device. If ``a`` is already on
        the device, no copy is performed.

    """
    if is_tensor(a):
        return a

    # numpy dont like tuple
    if isinstance(a, tuple):
        a = list(a)

    if not dtype:
        # automatic inference based of floatx and intx settings
        dtype = _infer_dtype(a)

    return np.asarray(a, dtype)


def copy(tensor):
    """Copy a tensor

    Args:
        tensor (ndarray): tensor to copy
    Returns
        ndarray: copied tensor
    """
    return np.copy(tensor)


def zeros(shape, dtype=floatx()):
    """Returns a new Tensor of given shape and dtype, filled with zeros.
    Args:
        shape (int or tuple of ints): Dimensionalities of the array.
        dtype: Data type specifier.

    Returns:
        ndarray: An tensor filled with zeros.

    """
    return np.zeros(shape, dtype=dtype)


def ones(shape, dtype=floatx()):
    """Returns a new Tensor of given shape and dtype, filled with ones.
    Args:
        shape (int or tuple of ints): Dimensionalities of the array.
        dtype: Data type specifier.

    Returns:
        ndarray: An tensor filled with zeros.

    """
    return np.ones(shape, dtype=dtype)


def fill(shape, fill_value):
    """Returns a new Tensor of given shape and dtype, filled with the provided
    value.

    Args:
        shape (int or tuple of ints): Dimensionalities of the array.
        fill_value (int): The value to fill with.
        dtype: Data type specifier.

    Returns:
        ndarray: An tensor filled with zeros.

    """
    return np.full(shape, fill_value)


def normal(shape, mean=0.0, stddev=1.0):
    """Draw random samples from a normal (Gaussian) distribution.
    Args:
        shape (int or tuple of ints): Dimensionalities of the array.
        mean (float): Mean value. Default to 0.0.
        dev (float): Standard deviations. Default to 1.0.

    Returns:
        ndarray: Drawn samples from the parameterized normal distribution.

    """
    return np.random.normal(loc=mean, scale=stddev, size=shape)


def range(start, stop=None, step=1, dtype=intx()):
    """Creates a sequence of numbers.
    Creates a sequence of numbers that begins at `start` and extends by
    increments of `delta` up to but not including `stop`.

    Args:
        start (int): Initial value. Optional. Defaults to 0
        stop (int, optional): End value.
        delta (int, optional): Spacing between values.  Defaults to 1.
        dtype (str, optional): Tensor tyoe. Defaults to intx().

    Returns:
        Tensor: Tensor that contains the requested range.
    """
    return np.arange(start, stop=stop, step=step, dtype=dtype)


# - Reduce -
def prod(tensor, axis=None, keepdims=False):
    """Returns the product of an array along a given axis.

    Args:
        tensor (ndarray): Array to take the maximum.
        axis (int): Along which axis to take the maximum. The flattened array
            is used by default. Defaults to None.
        dtype: Data type specifier.
        keepdims (bool): If ``True``, the axis is kept as an axis of
        size one. Default to False.

    Returns:
        ndarray: The maximum of ``tensor``, along the axis if specified.
    """
    return np.prod(tensor, axis=axis, keepdims=keepdims)


def max(tensor, axis=None, keepdims=False):
    """Returns the maximum of an array or the maximum along an axis.

    Note::
       When at least one element is NaN, the corresponding min value will be
       NaN.

    Args:
        tensor (ndarray): Array to take the maximum.
        axis (int): Along which axis to take the maximum. The flattened array
            is used by default. Defaults to None.
        dtype: Data type specifier.
        keepdims (bool): If ``True``, the axis is kept as an axis of
        size one. Default to False.

    Returns:
        ndarray: The maximum of ``tensor``, along the axis if specified.
    """
    return np.amax(tensor, axis=axis, keepdims=keepdims)


def min(tensor, axis=None, keepdims=False):
    """Returns the minimum of an array or the maximum along an axis.

    Note::
       When at least one element is NaN, the corresponding min value will be
       NaN.

    Args:
        tensor (ndarray): Array to take the maximum.
        axis (int): Along which axis to take the maximum. The flattened array
            is used by default. Defaults to None.
        keepdims (bool): If ``True``, the axis is kept as an axis of
        size one. Default to False.

    Returns:
        ndarray: The maximum of ``tensor``, along the axis if specified.
    """

    return np.amin(tensor, axis=axis, keepdims=keepdims)


def sum(tensor, axis=None, keepdims=False):
    """Returns the sum of an array along given axes.

    Args:
        tensor (ndarray): Array to sum reduce.
        axis (int or sequence of ints): Axes along which the sum is taken.
        keepdims (bool): If ``True``, the specified axes are remained as axes
        of length one.

    Returns:
        ndarray: The sum of ``tensor``, along the axis if specified.
    """
    return np.sum(tensor, axis=axis, keepdims=keepdims)


def mean(tensor, axis=None, keepdims=False):
    """Returns the sum of an array along given axes.

    Args:
        tensor (ndarray): Array to mean reduce.
        axis (int or sequence of ints): Axes along which the sum is taken.
        keepdims (bool): If ``True``, the specified axes are remained as axes
        of length one.

    Returns:
        ndarray: The mean of ``tensor``, along the axis if specified.
    """
    return np.mean(tensor, axis=axis, keepdims=keepdims)


def sqrt(tensor):
    """Computes element-wise square root of the input tensor.

    Args:
        tensor (ndarray): tensor

    Returns:
        tensor: square root of the input tensor.
    """
    return np.sqrt(tensor)


# - Manipulation -


def reverse(tensor, axis):
    """Reverses specific dimensions of a tensor.

    Args:
        tensor (tensor): tensor to reverse
        axis (tensor): axis or tuple of axis
    """
    return np.flip(tensor, axis)


def roll(tensor, shift, axis):
    """Rolls the elements of a tensor along an axis.

    Args:
        tensor (tensor): tensor to roll
        shift (tensor): offset to roll
        axis (tensor): axis to shift by

    Returns:
        [type]: [description]
    """
    return np.roll(tensor, shift, axis)


def assign(dst_tensor, values, slices):
    """ assign values in tensor at the position specified by the slices.

    Args:
        dst_tensor (ndarray): Target tensor
        values (ndarray): Tensor containing the values to assign.
        slices (tuple): slices that defines where to put the values.
    Returns:
        ndarray: tensor with the assigned values.

    """
    dst_tensor[slices] = values
    return dst_tensor


def tile(tensor, reps):
    """Construct a tensor by repeating tensor the number of times given by reps.
    Args:
        tensor (cupy.ndarray): Tensor to transform.
        reps (int or tuple): The number of repeats.
    Returns:
        ndarray: Transformed tensor with repeats.
    """
    return np.tile(tensor, reps)


def concatenate(tup, axis=0):
    """Joins tensors along a given axis.
    Args:
        tup (sequence of arrays): Tensors to be joined. Tensors must have the
        same cardinality except for the specified axis.

        axis (int or None): Axis to joint along. If None, tensors are
        flattened before use. Default is 0.
    Returns:
        ndarray: Joined array.
    """
    return np.concatenate(tup, axis=axis)


# - Utils -


def transpose(a):
    "Transpose the tensor"
    return np.transpose(a)


def cast(tensor, dtype):
    """Cast

    Args:
        tensor (Tensor): tensor to cast.
        dtype (str): type to cast. Usually floatx or intx
    """

    if isinstance(tensor, float) or isinstance(tensor, int):
        if dtype in ['float', 'float32']:
            return float(tensor)
        elif dtype == ['int', 'int32']:
            return int(tensor)
        else:
            raise ValueError("can't cast scalar", type(tensor), "to dype",
                             dtype)

    return tensor.astype(dtype)


def dtype(tensor):
    """"Returns the dtype of a tensor as a string.

    Args:
        tensor (tensor): Tensor

    Returns:
        str: type of the tensor as string. e.g int32.
    """
    return tensor.dtype.name


def flatten(tensor):
    """Returns a copy of the tensor flatten into one dimension.

    Args:
        tensor (ndarray): tensor to flatten
    Returns:
        ndarray: flattened tensor
    """
    return tensor.flatten()


def as_numpy_array(tensor):
    """Convert tensor to a numpy array.

    Useful when interfacing with other librairies to have a unified input
    to them.

    Args:
        tensor (ndarray): tensor to convert
    Returns:
        numpy.ndarray: Tensor as numpy array
    """
    return np.asarray(tensor)


def reshape(tensor, shape):
    "reshape tensor"
    return np.reshape(tensor, shape)


def is_tensor(a):
    "check if the given object is a tensor"
    return isinstance(a, np.ndarray)


def tensor_equal(tensor1, tensor2):
    """True if two tensors are exactly equals

    Args:
        tensor1 (ndarray): tensor to compare.
        tensor2 (ndarray): tensor to compare.

    Returns:
        Bool: True if exactly equal, False otherwise.

    """
    return np.array_equal(tensor1, tensor2)


def assert_near(a, b, absolute_tolerance=0, relative_tolerance=0):
    """
    Returns True if two arrays are element-wise equal within a tolerance.

    Args:
        a (ndarray): Tensor with a last dimension of at least k size.
        b (ndarray): Tensor with a last dimension of at least k size.
        absolute_tolerance (float): Default to 0
        relative_tolerance (float): Default to 0

    Returns:
        bool: True if the two arrays are equal within the given tolerance; False otherwise.

    Note:
        This function return True if the following equation is satisfied:
        `absolute(a - b) <= (absolute_tolerance + relative_tolerance * absolute(b))`  # noqa

    """

    return np.allclose(a, b, atol=absolute_tolerance, rtol=relative_tolerance)


# - Math -


def dot(t1, t2):
    """Return the dot product of two arrays

    Args:
        t1 (ndarray): Left tensor
        t2 (ndarray): Right tensor

    Return:
        ndarray: tensor containing the dot product
    """
    return np.dot(t1, t2)


def add(tensor1, tensor2):
    """Add two tensors

    Args:
        tensor1 (ndarray): Left tensor.
        tensor2 (ndarray): right tensor.
    """
    return np.add(tensor1, tensor2)


def subtract(tensor1, tensor2):
    """Substract two tensors

    Args:
        tensor1 (ndarray): Left tensor.
        tensor2 (ndarray): right tensor.
    """
    return np.subtract(tensor1, tensor2)


def multiply(tensor1, tensor2):
    """multiply two tensors

    Args:
        tensor1 (ndarray): Left tensor.
        tensor2 (ndarray): right tensor.
    """
    return np.multiply(tensor1, tensor2)


def divide(numerator, denominator):
    """divide a tensor by another

    Args:
        tensor1 (ndarray): numerator tensor.
        tensor2 (ndarray): denominator tensor.
    """
    return np.divide(numerator, denominator)


def mod(numerator, denominator):
    """Compute the reminder of the divisin of a tensor by another

    Args:
        tensor1 (ndarray): numerator tensor.
        tensor2 (ndarray): denominator tensor.
    """
    return np.mod(numerator, denominator)


def clip(tensor, min_val=0, max_val=None):
    """Clips the values of a tensor to a given interval. For example,
    if an interval of [0, 1] is specified, values smaller than 0 become 0,
    and values larger than 1 become 1.

    Efficient version of ``max(min(a, max_val), min_val)``

    Args:
        tensor (ndarray): The input Tensor.

        min_val (scalar, ndarray ): The left side of the interval.
        Defaults to 0.

        max_val (scalar, ndarray or None): The right side of the interval. When
        None ignored. Defaults to None.

    Returns:
        ndarray: Clipped tensor.
    """
    return np.clip(tensor, a_min=min_val, a_max=max_val)


def abs(tensor):
    "Calculate the absolute value element-wise."
    return np.absolute(tensor)


def broadcasted_norm(tensor):
    "Norm broadcasted accross dimensions"
    return np.sum(np.abs(tensor)**2, axis=-1)**0.5


def norm(tensor, ord=None, axis=None, keepdims=False):
    """Return one of eight different matrix norms, or one of an infinite
    number of vector norms (described below), depending on the value of
    the `ord` parameter.

    Args:
        tensor (ndarray): Array to take the norm from. If ``axis`` is None,
        the ``tensor`` must be 1D or 2D.

        ord (non-zero int, inf, -inf, 'fro'): Norm type.
        See `numpy.linalg.norm` for explanation.

        axis (int, 2-tuple of ints, None): `axis` along which the norm is
        computed.

        keepdims (bool): If this is set ``True``, the axes which are normed
        over are left in the resulting tensor with a size of one.

    Returns:
        ndarray: Norm of the tensor.
    """
    return np.linalg.norm(tensor, ord=ord, axis=axis, keepdims=keepdims)


# - Randomness -


def randint(low, high=None, shape=None, dtype='l'):
    """Returns a scalar or an array of integer values over [low, high)

    Args:
        low (int): If high is None, it is the upper bound of the
        interval and the lower bound is set to 0. if high is set, it is the
        lower bound of the interval.

        high (int, optional):Upper bound of the interval. Defaults to None.

        shape (None or int or tuple of ints, optional): The shape of returned
        value. Defaults to None.

        dtype (str, optional): dtype: Data type specifier. Defaults to 'l'.

    Returns:
        int or ndarray of ints: If size is None, it is single integer
        sampled. If size is integer, it is the 1D-array of length size
        element. Otherwise, it is the array whose shape specified by size.
    """
    return np.random.randint(low, high=high, size=shape, dtype=dtype)


def shuffle(tensor, axis=0):
    """Shuffle a tensor along a given axis. Other axis remain
    in place.


    Args:
        tensor (ndarray): tensor to shuffle.
        axis (int, optional): axis to shuffle on. Default to 0.

    Returns:
        tensor: shuffled tensor
    """

    # ! we must return the tensor because other backend don't do shuffle
    # ! in place.
    # FIXME move the rng to a global one to avoid allocation
    rng = np.random.default_rng()
    rng.shuffle(tensor, axis=axis)
    return tensor


def full_shuffle(tensor):
    """Shuffle a tensor along all of its axis

    Args:
        tensor (ndarray): tensor to shuffle.

    Returns:
        Tensor:shuffled tensor

    """
    # ! we must return the tensor because other backend don't do shuffle
    # ! in place.
    rng = np.random.default_rng()
    for idx in range(len(tensor.shape)):
        rng.shuffle(tensor, axis=idx)
    return tensor


# - Indexing -


def take(tensor, indices, axis=None):
    """Takes elements of a Tensor at specified indices along a specified axis

    Args:
        tensor (ndarray): Tensor to extract elements from.

        indices (int or array-like): Indices of elements that this function
        takes.

        axis (int, optional): The axis along which to select indices from.
        The flattened input is used by default. Defaults to None.

        out (ndarray. Optional): Output array. If provided, it should be of
            appropriate shape and dtype. Defaults to None.
    Returns:
        ndarray: Tensor containing the values from the specified indices.
    """
    return np.take(tensor, indices, axis=axis)


def top_k_indices(tensor, k):
    """
    Finds the indices of the k largest entries alongside an axis.

    Args:
        tensor (ndarray): Tensor with a last dimension of at least k size.

        k (i): number of elements to return.

    """
    k = -k  # reverse to get top elements

    # we do partition and then sort to maximizes speed and have an avg
    # complexity of O(k log k).
    idxs = np.argpartition(tensor, k)[k:]

    # ! mind the - argsort return in increasing order we want largest
    return idxs[np.argsort(-tensor[idxs])]


def bottom_k_indices(tensor, k):
    """
    Finds the indices of the k smallest entries alongside an axis.

    Args:
        tensor (ndarray): Tensor with a last dimension of at least k size.

        k (i): number of elements to return.
    """
    idxs = np.argpartition(tensor, k)[:k]
    return idxs[np.argsort(tensor[idxs])]


def unique_with_counts(tensor):
    """Finds unique elements and return them along side their position and
    counts.

    Args:
        tensor (Tensor): 1D tensor to analyze.

    Returns:
        values (Tensor): unique values founded
        indexes (Tensor): index of the value sorted
        counts (Tensor): Tensor containing the count for each value.
    """
    return np.unique(tensor, return_index=True, return_counts=True)


# - Statistical -


def bincount(tensor, weights=None, minlength=0):
    """Count number of occurrences of each value in array of non-negative ints.

    Args:
        tensor (ndarray): Input tensor.

        weights (cupy.ndarray): Weights tensor which has the same shape as
        tensor``. Default to None.

        minlength (int): A minimum number of bins for the output array.

    Returns:
        ndarray: The result of binning the input tensor. The length of
        output is equal to ``max(max(x) + 1, minlength)``.

    """
    return np.bincount(tensor, weights=weights, minlength=minlength)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.utils import box, unbox
from evoflow.backend import tensor
from evoflow.backend import tensor_equal


def test_box_unbox():
    vals = ['a', [1, 2, 3]]

    for val in vals:
        assert unbox(box(val)) == val


def test_box_unbox_tensor():
    val = tensor([1, 2, 3])

    assert tensor_equal(unbox(box(val)), val)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from termcolor import cprint
from evoflow import backend as B
from evoflow.ops import Reverse1D, Reverse2D, Reverse3D
from evoflow.population import uniform_population


def test_ND():
    "test various tensor size random"

    TEST_INPUTS = [
        [Reverse1D, (2, 4), 0.5],
        [Reverse2D, (2, 4, 4), (0.5, 0.5)],
        [Reverse3D, (2, 4, 4, 4), (0.5, 0.5, 0.5)],
    ]

    for inputs in TEST_INPUTS:
        OP = inputs[0]
        pop_shape = inputs[1]
        max_reverse_probability = inputs[2]
        population_fraction = 1
        population = B.randint(0, 1024, pop_shape)

        # eager
        RM = OP(population_fraction=population_fraction,
                max_reverse_probability=max_reverse_probability)

        population = RM(population)
        assert B.is_tensor(population)
        assert population.shape == pop_shape

        # graph
        RM = OP(population_fraction=population_fraction,
                max_reverse_probability=max_reverse_probability)

        population = RM._call_from_graph(population)
        assert B.is_tensor(population)
        assert population.shape == pop_shape


def test_1D_tensor_values_maintained():
    POPULATION_SHAPE = (6, 8)
    # need a uniform population otherwise its not testable.
    population = uniform_population(POPULATION_SHAPE)
    population_fraction = 0.5
    max_reverse_probability = 0.2

    cprint(population, 'cyan')
    population = Reverse1D(population_fraction,
                           max_reverse_probability,
                           debug=1)(population)

    cprint(population, 'yellow')
    for chrm in population:
        _, _, count = B.unique_with_counts(chrm)
        assert B.max(count) == 1



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from termcolor import cprint
from evoflow.backend.common import intx


def test_take(backends):
    data = [
        # tensor, indices to take, expected
        [[0, 1, 2, 3, 4, 5], [5, 1, 3], [5, 1, 3]],
    ]
    for d in data:
        for B in backends:
            tensor = B.tensor(d[0])
            idxs = B.tensor(d[1])
            expected = B.tensor(d[2])

            result = B.take(tensor, idxs)
            assert B.tensor_equal(result, expected)


def test_topk(backends):
    data = [
        # tensor, k, expected
        [[1, 10, 2, 20, 3, 30], 3, [5, 3, 1]],
        [[-1, 10, 2, -20, 3, 30], 3, [5, 1, 4]],
    ]
    for d in data:
        for B in backends:
            tensor = B.tensor(d[0])
            k = d[1]
            expected = B.tensor(d[2])

            result = B.top_k_indices(tensor, k)
            print(result)
            assert B.tensor_equal(result, expected)


def test_bottom_k_indices(backends):
    data = [
        # tensor, k, expected
        [[1, 10, 2, 20, 3, 30], 3, [0, 2, 4]],
        [[-1, 10, 2, -20, 3, 30], 3, [3, 0, 2]],
    ]
    for d in data:
        for B in backends:
            cprint('intx dtype:%s' % intx(), 'magenta')
            tensor = B.tensor(d[0])
            cprint('tensor dtype:%s' % B.dtype(tensor), 'green')
            k = d[1]
            expected = B.tensor(d[2])
            cprint('expected dtype:%s' % B.dtype(expected), 'cyan')

            result = B.bottom_k_indices(tensor, k)
            cprint('expected dtype:%s' % B.dtype(result), 'yellow')

            assert B.tensor_equal(result, expected)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.config import floatx, set_floatx, intx, set_intx


def test_tensor(backends):
    data = [[[1, 2, 3, 4], intx()], [[0.1, 2, 3], floatx()],
            [[3, 0.1, 3], floatx()]]
    for d in data:
        for B in backends:
            tensor = B.tensor(d[0])
            assert B.dtype(tensor) == d[1]


def test_set_floatx_tensor(backends):
    data = [[1.1, 2, 3, 4]]
    for d in data:
        for B in backends:
            set_floatx('float16')
            tensor = B.tensor(d)
            assert B.dtype(tensor) == floatx()
            set_floatx('float32')


def test_set_intx_tensor(backends):
    data = [[1, 2, 3, 4]]
    for d in data:
        for B in backends:
            set_intx('uint8')
            tensor = B.tensor(d)
            assert B.dtype(tensor) == intx()
            set_intx('int32')


def test_set_tensor_dtype(backends):
    data = [[1, 2, 3, 4]]
    for d in data:
        for B in backends:
            tensor = B.tensor(d, 'int16')
            assert B.dtype(tensor) == 'int16'


def test_copy(backends):
    data = [[1, 2, 3, 4]]
    for d in data:
        for B in backends:
            tensor = B.tensor(d)
            tensor2 = B.copy(tensor)

            # same before modification
            assert B.tensor_equal(tensor, tensor2)

            # different after modification
            tensor2 = B.assign(tensor2, 42, slice(2, 3))
            print(tensor2)
            assert not B.tensor_equal(tensor, tensor2)


def test_zero(backends):
    shape = (10, 10)

    for B in backends:
        tensor = B.zeros(shape)
        assert tensor.shape == shape
        assert not B.sum(tensor)


def test_ones(backends):
    shape = (10, 10)

    for B in backends:
        tensor = B.ones(shape)
        assert tensor.shape == shape
        assert B.sum(tensor) == 100


def test_fill(backends):
    shape = (10, 10)

    for B in backends:
        tensor = B.fill(shape, 5)
        assert tensor.shape == shape
        assert B.sum(tensor) == 500



from evoflow.backend import set_floatx, floatx, set_intx, intx


def test_set_floatx():
    for dtype in ['float16', 'float32', 'float64']:
        set_floatx(dtype)
        assert floatx() == dtype
        set_floatx('float32')


def test_set_intx():
    for dtype in ['int8', 'uint8', 'int16', 'uint16', 'int32', 'int64']:
        set_intx(dtype)
        assert intx() == dtype
        set_intx('int32')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from evoflow.io import print_debug


class SelectionStrategy(object):
    """Selection Strategy base classe.
    To be implemented by subclasses:
    * `call()`: Contains the logic for selecting which chromesome to keep.
    """
    def __init__(self, name, **kwargs):
        self.name = name
        self.debug = kwargs.get('debug', False)

    @abc.abstractmethod
    def call(self, fitness_function, current_population, evolved_population):
        """Select new population.

        Args:
            fitness_function (function): User provided function that return
            the fitness value for each chromosome of a population as a Tensor.

            current_population: Tensor containing the population prior to
            evolution.

            evolved_population: Tensor containing the population after the
            evolution

        """
        NotImplementedError('Must be implemented in subclasses.')

    def __call__(self, fitness_function, current_population,
                 evolved_population):
        """Select new population.

        Args:
            fitness_function (function): User provided function that return
            the fitness value for each chromosome of a population as a Tensor.

            current_population: Tensor containing the population prior to
            evolution.

            evolved_population: Tensor containing the population after the
            evolution

        Returns
            [new_populations, fitness_values]

        """
        return self.call(fitness_function, current_population,
                         evolved_population)

    def print_debug(self, *msg):
        "output debug message"
        if self.debug:
            name = self.__class__.__name__
            print_debug(name, msg)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.engine import OP
from evoflow import backend as B


class Shuffle(OP):
    def __init__(self, population_fraction, **kwargs):
        """Shuffle genes within the chromsome.

        Args:
            population_fraction (float): How many chromosomes
            should have a cross-over.

            debug (bool, optional): print debug information and function
            returns additional data.

        Returns:
            tensor: population with shuffled genes.

        See:
            https://en.wikipedia.org/wiki/Crossover_(genetic_algorithm)
        """

        if not (0 < population_fraction <= 1.0):
            raise ValueError("num_crossover_fraction must be in [0, 1]")

        self.population_fraction = population_fraction
        super(Shuffle, self).__init__(**kwargs)

    def call(self, populations):
        results = []
        for population in populations:
            results.append(self.compute(population))
        return results

    def compute(self, population):

        # shuffle to make sure  don't hit the same everytime
        if not self.debug:
            population = B.shuffle(population)

        # how many chromosomes to shuffle
        num_shuffle = int(population.shape[0] * self.population_fraction)

        shuffled_population = population[:num_shuffle]
        shuffled_population = B.full_shuffle(shuffled_population)
        self.print_debug("shuffled population", shuffled_population.shape)

        # recompose with non shuffled population
        shuffled_population = B.concatenate(
            [shuffled_population, population[num_shuffle:]])

        return shuffled_population


if __name__ == '__main__':
    from copy import copy
    GENOME_SHAPE = (6, 4, 4)
    population = B.randint(0, 256, GENOME_SHAPE)
    population_fraction = 0.5
    max_crossover_size_fraction = (0.5, 0.5)
    print(population.shape)
    original_population = copy(population)
    population = Shuffle(population_fraction, debug=True)(population)

    # diff matrix
    diff = B.clip(abs(population - original_population), 0, 1)
    print(diff)



from evoflow.population import randint_population, uniform_population
import evoflow.backend as B


def test_randintpop():
    shape = (100, 100, 10)
    pop = randint_population(shape, 42, 1)
    assert pop.shape == shape
    assert B.max(pop) == 42
    assert B.min(pop) == 1


def test_uniformpop():
    shape = (100, 100)
    pop = uniform_population(shape)
    assert pop.shape == shape
    assert B.max(pop) == 99
    assert B.min(pop) == 0

    for chrm in pop:
        _, _, count = B.unique_with_counts(chrm)
        assert B.max(count) == 1



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_roll(backends):
    inputs = [1, 2, 3]
    expected = [3, 1, 2]

    for B in backends:
        tensor = B.tensor(inputs)
        result = B.roll(tensor, 1, axis=0)
        expected_tensor = B.tensor(expected)
        assert B.tensor_equal(result, expected_tensor)


def test_roll_2d(backends):
    inputs = [[1, 2, 3], [1, 2, 3]]
    expected = [[3, 1, 2], [3, 1, 2]]

    for B in backends:
        tensor = B.tensor(inputs)
        result = B.roll(tensor, 1, axis=1)
        expected_tensor = B.tensor(expected)
        assert B.tensor_equal(result, expected_tensor)


def test_assign(backends):
    inputs = [[11, 12, 13], [21, 22, 23], [31, 32, 33]]
    values = [[1, 1], [1, 1]]
    slices = (slice(1, 3), slice(1, 3))

    expected = [[11, 12, 13], [21, 1, 1], [31, 1, 1]]

    for B in backends:
        tval = B.tensor(values)
        tensor = B.tensor(inputs)
        result = B.assign(tensor, tval, slices)
        expected_tensor = B.tensor(expected)
        assert B.tensor_equal(result, expected_tensor)


def test_concatenate(backends):
    inputs = [1, 2, 3]
    inputs2 = [4, 5, 6]
    expected = [1, 2, 3, 4, 5, 6]
    for B in backends:
        tensor = B.tensor(inputs)
        tensor2 = B.tensor(inputs2)
        result = B.concatenate([tensor, tensor2])
        expected_tensor = B.tensor(expected)
        assert B.tensor_equal(result, expected_tensor)



_INTX = 'int32'
_FLOATX = 'float32'
_BACKEND = None


def floatx():
    """Returns the default float type, as a string.

    E.g. `'float16'`, `'float32'`, `'float64'`.

    Returns:
        str: the current default float type.

    Example:
    >>> B.floatx()
    'float32'
    """
    return _FLOATX


def set_floatx(value):
    """Sets the default float type.

    Note: It is not recommended to set this to float16 for training,
    as this will likely cause numeric stability issues

    Args:
        value (str): `'float16'`, `'float32'`, or `'float64'`.

    Example:
    >>> evoflow.backend.floatx()
    'float32'
    >>> evoflow.backend.set_floatx('float64')
    >>> evoflow.backend.floatx()
    'float64'
    >>> evoflow.backend.set_floatx('float32')
    Raises:
        ValueError: In case of invalid value.
    """
    global _FLOATX
    if value not in {'float16', 'float32', 'float64'}:
        raise ValueError('Unknown floatx type: ' + str(value))
    _FLOATX = str(value)


def intx():
    """Returns the default int type, as a string.

    E.g. `'int8'`, `'unit8'`, `'int32'`.

    Returns:
        str: the current default int type.

    Example:
    >>> B.intx()
    'int32'
    """
    return _INTX


def set_intx(value):
    """Sets the default int type.

    Args:
        value (str): default int type to use.
        One of `{'int8', 'uint8', 'int16', 'uint16', 'int32', 'int64'}`

    Example:
    >>> evoflow.backend.intx()
    'int32'
    >>> evoflow.set_intx('uint8')
    >>> evoflow.backend.intx()
    'uint8'
    >>> evoflow.intx('float132')
    Raises:
        ValueError: In case of invalid value.
    """
    global _INTX
    if value not in {'int8', 'uint8', 'int16', 'uint16', 'int32', 'int64'}:
        raise ValueError('Unknown intx type: ' + str(value))
    _INTX = str(value)


def get_backend():
    "Return the backend used"
    return _BACKEND


def set_backend(name):
    """Set Geneflow backend to be a given framework

    Args:
        name(str): Name of the backend. {cupy, numpy, tensorflow}. Default
        to tensorflow.

    See:
        `load_backend.py` for the actual loading code.
    """
    global _BACKEND
    if name not in {'cupy', 'numpy', 'tensorflow'}:
        raise ValueError('Unknown backend: ' + str(name))

    _BACKEND = name



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import evoflow.backend as B
from evoflow.engine import FitnessFunction


class Sum(FitnessFunction):
    def __init__(self, max_sum_value=None, **kwargs):
        """Compute the sum of the chromosomes as fitness values.

        Args:
            expected_max_sum_value (int, optional): What is the maximum value
            the sum per chromosome will reach. Used to normalize the fitness
            function between 0 and 1 if specified. Defaults to None.

        Note:
            This fitness function is used to solve the MAXONE problem.

        """
        super(Sum, self).__init__(**kwargs)
        if max_sum_value:
            self.max_sum_value = B.tensor(max_sum_value)
        else:
            self.max_sum_value = None

    def call(self, population):
        flat_pop = self._flatten_population(population)
        scores = B.sum(flat_pop, axis=-1)
        if self.max_sum_value:
            return scores / self.max_sum_value
        else:
            return scores



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.backend.common import floatx, intx


def test_to_numpy(backends):
    inputs = [[1, 2, 3], [0.1, 0.2, 0.3], [1, 0.1, 0.2]]
    for ipt in inputs:
        for B in backends:
            tensor = B.tensor(ipt)
            numpy_arr = B.as_numpy_array(tensor)

            # check conversion type is consistent
            assert B.dtype(tensor) == numpy_arr.dtype.name

            # checking that converting back and forth works
            tensor_back = B.tensor(numpy_arr)
            assert B.tensor_equal(tensor, tensor_back)


def test_cast_dtype(backends):
    inputs = [1, 2, 3]
    for B in backends:
        tensor = B.tensor(inputs)
        tensor = B.cast(tensor, floatx())
        assert B.dtype(tensor) == floatx()
        tensor = B.cast(tensor, intx())
        assert B.dtype(tensor) == intx()


def test_reshape(backends):
    inputs = [[1, 2, 3], [4, 5, 6]]
    for B in backends:
        tensor = B.tensor(inputs)
        reshaped_tensor = B.reshape(tensor, 6)
        assert reshaped_tensor.shape == (6, )

        # check back and forth
        reshaped_tensor = B.reshape(reshaped_tensor, (2, 3))
        assert B.tensor_equal(tensor, reshaped_tensor)


def test_is_tensor(backends):
    inputs = [1, 2, 3]
    for B in backends:
        tensor = B.tensor(inputs)
        assert B.is_tensor(tensor)
        assert not B.is_tensor(inputs)


def test_tensor_equal(backends):
    inputs = [1, 2, 3]
    same_inputs = [1, 2, 3]
    different_inputs = [1, 3, 2]

    for B in backends:
        tensor = B.tensor(inputs)
        same_tensor = B.tensor(same_inputs)
        different_tensor = B.tensor(different_inputs)
        assert B.tensor_equal(tensor, same_tensor)
        assert not B.tensor_equal(tensor, different_tensor)


def test_tensor_are_near(backends):
    inputs = [1, 2, 3]
    near_inputs = [1.1, 2.1, 3.1]
    different_inputs = [1, 3, 2]

    for B in backends:
        tensor = B.tensor(inputs)
        tensor = B.cast(tensor, floatx())
        near_tensor = B.tensor(near_inputs)
        different_tensor = B.tensor(different_inputs)
        different_tensor = B.cast(different_tensor, floatx())

        assert B.assert_near(tensor, near_tensor, absolute_tolerance=0.2)

        # those should fail
        assert not B.tensor_equal(tensor, near_tensor)
        assert not B.tensor_equal(tensor, different_tensor)


# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.ops import RandomMutations1D, RandomMutations2D
from evoflow.ops import RandomMutations3D
from evoflow import backend as B
from copy import copy
from termcolor import cprint


def test_binary_val_default_params():
    pop_shape = (6, 4)
    population = B.randint(0, 2, pop_shape)
    population = RandomMutations1D(max_gene_value=1, debug=1)(population)
    print(population)
    assert B.max(population) == 1
    assert not B.min(population)


def test_clip():
    pop_shape = (100, 100)
    population = B.randint(0, 100, pop_shape)
    population = RandomMutations1D(max_gene_value=100, debug=1)(population)
    assert B.max(population) <= 100
    assert not B.min(population)


def test_mutation2d_eager():
    pop_shape = (2, 4, 4)
    max_gene_value = 10
    min_gene_value = 0
    population_fraction = 1
    mutations_probability = (0.5, 0.5)
    min_mutation_value = 1
    max_mutation_value = 1

    population = B.randint(0, max_gene_value, pop_shape)

    # save original
    original_population = copy(population)
    cprint('[Initial genepool]', 'blue')
    cprint(original_population, 'blue')

    RM = RandomMutations2D(population_fraction=population_fraction,
                           mutations_probability=mutations_probability,
                           min_gene_value=min_gene_value,
                           max_gene_value=max_gene_value,
                           min_mutation_value=min_mutation_value,
                           max_mutation_value=max_mutation_value,
                           debug=True)
    population = RM(population)

    cprint('\n[Mutated genepool]', 'yellow')
    cprint(population, 'yellow')

    cprint('\n[Diff]', 'magenta')
    diff = population - original_population
    cprint(diff, 'magenta')

    assert B.is_tensor(population)
    assert population.shape == pop_shape
    assert B.max(diff) <= max_mutation_value
    for chromosome in diff:
        assert B.sum(chromosome) == 4


def test_mutation2d_graph_mode():
    "make sure the boxing / unboxing works in graph mode"
    pop_shape = (2, 4, 4)
    max_gene_value = 10
    min_gene_value = 0
    population_fraction = 1
    mutations_probability = (0.5, 0.5)
    min_mutation_value = 1
    max_mutation_value = 1

    population = B.randint(0, max_gene_value, pop_shape)

    RM = RandomMutations2D(population_fraction=population_fraction,
                           mutations_probability=mutations_probability,
                           min_gene_value=min_gene_value,
                           max_gene_value=max_gene_value,
                           min_mutation_value=min_mutation_value,
                           max_mutation_value=max_mutation_value,
                           debug=True)

    population = RM._call_from_graph(population)
    assert B.is_tensor(population)
    assert population.shape == pop_shape


def test_ND():
    "test various tensor size random"

    TEST_INPUTS = [
        [RandomMutations1D, (2, 4), 0.5],
        [RandomMutations2D, (2, 4, 4), (0.5, 0.5)],
        [RandomMutations3D, (2, 4, 4, 4), (0.5, 0.5, 0.5)],
    ]

    for inputs in TEST_INPUTS:
        OP = inputs[0]
        pop_shape = inputs[1]
        mutations_probability = inputs[2]

        max_gene_value = 10
        min_gene_value = 0
        population_fraction = 1
        min_mutation_value = 1
        max_mutation_value = 1

        population = B.randint(0, max_gene_value, pop_shape)

        # eager
        RM = OP(population_fraction=population_fraction,
                mutations_probability=mutations_probability,
                min_gene_value=min_gene_value,
                max_gene_value=max_gene_value,
                min_mutation_value=min_mutation_value,
                max_mutation_value=max_mutation_value)

        population = RM(population)
        assert B.is_tensor(population)
        assert population.shape == pop_shape

        # graph
        RM = OP(population_fraction=population_fraction,
                mutations_probability=mutations_probability,
                min_gene_value=min_gene_value,
                max_gene_value=max_gene_value,
                min_mutation_value=min_mutation_value,
                max_mutation_value=max_mutation_value)

        population = RM._call_from_graph(population)
        assert B.is_tensor(population)
        assert population.shape == pop_shape


def test_uniform_distribution():
    """check that every gene of the tensor are going to be flipped equally
    Note:
    # ! We need enough iterations and chromosomes to reduce collision
    # ! and ensure numerical stability
    """

    NUM_ITERATIONS = 1000
    GENOME_SHAPE = (20, 4, 4)
    population = B.randint(0, 1024, GENOME_SHAPE)
    population_fraction = 1
    crossover_probability = (0.5, 0.5)

    # each gene proba of being mutated 0.5*0.5 > 0.25
    # each chromosome proba of being mutated 1
    # => gene average hit rate: 1000 / (1/4)  ~250
    MIN_DIFF_BOUND = 200
    MAX_DIFF_BOUND = 300

    OP = RandomMutations2D(population_fraction, crossover_probability)

    # diff matrix
    previous_population = copy(population)
    population = OP(population)
    diff = B.clip(abs(population - previous_population), 0, 1)
    for _ in range(NUM_ITERATIONS - 1):
        previous_population = copy(population)
        population = OP(population)

        curr_diff = B.clip(abs(population - previous_population), 0, 1)
        # acumulating diff matrix
        diff += curr_diff

    print(curr_diff)

    for c in diff:
        print(c)
        print('mean', B.mean(c), 'min', B.min(c), 'max', B.max(c))
        assert B.min(c) > MIN_DIFF_BOUND
        assert B.max(c) < MAX_DIFF_BOUND
        assert MIN_DIFF_BOUND < B.mean(c) < MAX_DIFF_BOUND



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .op import OP
import evoflow.backend as B


class Input(OP):
    def __init__(self, shape, **kwargs):
        """Input op

        Args:
            shape (set of ints): what is the shape of the input to the
            evolutionary graph. Its of the form
            (num_chromosme, chromosome_dim_1, .... chromesome_dim_n)

            name (str, optional): Name of the OP. Should be unique in a graph
            (do not reuse the same name twice). It will be autogenerated
            if it isn't provided.
        """
        self.shape = shape
        super(Input, self).__init__(**kwargs)

    def get_output_shapes(self):
        "return the shape of tensor returned by the op"
        return self.shape

    def assign(self, chromosomes):
        """Assign concrete values to the input
        """
        chromosomes = B.tensor(chromosomes)

        if chromosomes.shape != self.shape:
            raise ValueError(
                'Incompatible input shape expected: %s - got: %s' %
                (self.shape, chromosomes.shape))
        self.chromosomes = chromosomes

    def call(self, unused):
        "make it compatible with other ops"
        return self.chromosomes

    def get(self):
        "shorthand version which is used inside our main loop"
        return self.chromosomes


# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import evoflow.backend as B
from evoflow.engine import SelectionStrategy


class SelectFittest(SelectionStrategy):
    "Select the fittest member of the population"

    def __init__(self, mode='max', **kwargs):
        """[summary]

        Args:
            mode (str, optional):  one of `{min', 'max'}`. In 'min' mode,
            the fitness function will select individual with the lowest fitness
            value; in 'max' mode it will select the one with the highest
            values. Defaults to 'max'.
        """
        if mode not in ['min', 'max']:
            raise ValueError('mode must be either max or min')

        self.mode = mode
        super(SelectFittest, self).__init__('select_fittest', **kwargs)

    def call(self, fitness_function, current_population, evolved_population):
        """Select the most fit individuals from the combined current and
        evolved population.

        Args:
            fitness_function (function): User provided function that return
            the fitness value for each chromosome of a population as a Tensor.

            current_population: Tensor containing the population prior to
            evolution.

            evolved_population: Tensor containing the population after the
            evolution
        """
        population_size = current_population.shape[0]
        merged_population = B.concatenate(
            [current_population, evolved_population])

        # fitness computation
        fitness_scores = fitness_function(merged_population)
        metrics = fitness_function.get_metrics()

        # population selection
        if self.mode == 'max':
            indices = B.top_k_indices(fitness_scores, k=population_size)
        else:
            indices = B.bottom_k_indices(fitness_scores, k=population_size)

        selected_pop = B.take(merged_population, indices, axis=0)

        return selected_pop, B.take(fitness_scores, indices, axis=0), metrics



from evoflow.config import floatx, intx


def _infer_dtype(a):
    "infers what tensor.dtype to use for a given variable during conversion"
    if isinstance(a, int):
        dtype = intx()
    elif isinstance(a, float):
        dtype = floatx()
    elif isinstance(a, list):
        have_float = False
        for v in a:
            if isinstance(v, float):
                have_float = True
        if have_float:
            dtype = floatx()
        else:
            dtype = intx()
    else:
        raise ValueError("can't cast type:", type(a), 'to a tensor')
    return dtype



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.engine import OP
from evoflow import backend as B
from termcolor import cprint


class RandomMutations(OP):
    def __init__(self,
                 population_fraction,
                 mutations_probability,
                 min_gene_value=0,
                 max_gene_value=None,
                 min_mutation_value=1,
                 max_mutation_value=1,
                 **kwargs):
        """Perform random mutations

        Args:
            population_fraction (float): What fraction of the population should
            be affected by the mutations.

            mutations_probability (list(float)): What fraction of the
            genes should be affected by the mutations.

            min_gene_value = (int, optional): min value a gene can take.
            Defaults to 0.

            max_gene_value = (int, optional): max value a gene can take.
            Defaults to None.

            min_mutation_value (int): Minimal value the mutation
            can take, can be negative. Defaults to 0.

            max_mutation_value (int): Max value the mutations can
            take. Defaults to 1.

            debug (bool, optional): Defaults to False.
        """

        if not (0 < population_fraction <= 1.0):
            raise ValueError("population_fraction must be in ]0. 1]")

        for val in mutations_probability:
            if not (0 < val <= 1.0):
                raise ValueError(
                    "mutations_probability values must be between ]0. 1]")

        if min_gene_value and max_gene_value and min_gene_value >= max_gene_value:  # noqa
            raise ValueError("min_gene_value must < than max_gen_value")

        # ! don't remove the _num_ qualifier. It might feel it is unecessary
        # ! but we want consistent naming accross ops and crossover ops need it
        self.population_fraction = population_fraction
        self.mutations_probability = mutations_probability
        self.min_gene_value = min_gene_value
        self.max_gene_value = max_gene_value
        self.min_mutation_value = min_mutation_value
        self.max_mutation_value = max_mutation_value
        super(RandomMutations, self).__init__(**kwargs)

    def call(self, populations):
        """ Create the mask use to generate mutations

        Args:
            population_shape (list): population tensor shape.
        Returns:
            tensor: mask

        Generation works by:
        1. creating a slice that contains the mutation
        2. Inserting it into the mask
        3. Shuffle the mask in every dimension to distribute them
        """
        results = []

        for population in populations:

            affected_population = int(population.shape[0] *
                                      self.population_fraction)

            # Build sub tensors & slices by iterating through tensor dimensions
            sub_tensor_shape = [affected_population]
            slices = [slice(0, affected_population)]
            for idx, pop_size in enumerate(population.shape[1:]):
                midx = idx - 1  # recall dim1 are genes.
                tsize = int(pop_size * self.mutations_probability[midx])
                sub_tensor_shape.append(tsize)
                slices.append(slice(0, tsize))
            slices = tuple(slices)
            self.print_debug("sub_tensor_shape", sub_tensor_shape)

            # drawing mutations
            mutations = B.randint(self.min_mutation_value,
                                  self.max_mutation_value + 1,
                                  shape=sub_tensor_shape)
            # blank mask
            mask = B.zeros(population.shape, dtype=B.intx())

            # add mutations
            mask = B.assign(mask, mutations, slices)

            # shuffle mask every axis
            mask = B.full_shuffle(mask)

            # mutate
            population = population + mask

            # normalize
            if self.max_gene_value or self.min_gene_value:
                self.print_debug("min_gen_val", self.min_gene_value)
                self.print_debug("max_gen_val", self.max_gene_value)

                population = B.clip(population,
                                    min_val=self.min_gene_value,
                                    max_val=self.max_gene_value)
            results.append(population)
        return results


class RandomMutations1D(RandomMutations):
    def __init__(self,
                 population_fraction=0.9,
                 mutations_probability=0.5,
                 min_gene_value=0,
                 max_gene_value=None,
                 min_mutation_value=1,
                 max_mutation_value=1,
                 **kwargs):
        if not isinstance(mutations_probability, float):
            raise ValueError('mutations_probability must be a float')

        super(RandomMutations1D,
              self).__init__(population_fraction,
                             mutations_probability=[mutations_probability],
                             min_gene_value=min_gene_value,
                             max_gene_value=max_gene_value,
                             min_mutation_value=min_mutation_value,
                             max_mutation_value=max_mutation_value,
                             **kwargs)


class RandomMutations2D(RandomMutations):
    def __init__(self,
                 population_fraction=0.9,
                 mutations_probability=(0.5, 0.5),
                 min_gene_value=0,
                 max_gene_value=None,
                 min_mutation_value=1,
                 max_mutation_value=1,
                 **kwargs):
        if len(mutations_probability) != 2:
            raise ValueError('mutations_probability must be of form (x, y)')

        super(RandomMutations2D,
              self).__init__(population_fraction=population_fraction,
                             mutations_probability=mutations_probability,
                             min_gene_value=min_gene_value,
                             max_gene_value=max_gene_value,
                             min_mutation_value=min_mutation_value,
                             max_mutation_value=max_mutation_value,
                             **kwargs)


class RandomMutations3D(RandomMutations):
    def __init__(self,
                 population_fraction=0.9,
                 mutations_probability=(0.5, 0.5, 0.5),
                 min_gene_value=0,
                 max_gene_value=None,
                 min_mutation_value=0,
                 max_mutation_value=1,
                 **kwargs):
        if len(mutations_probability) != 3:
            raise ValueError('mutations_probability must be of form (x, y, z)')

        super(RandomMutations3D,
              self).__init__(population_fraction=population_fraction,
                             mutations_probability=mutations_probability,
                             min_gene_value=min_gene_value,
                             max_gene_value=max_gene_value,
                             min_mutation_value=min_mutation_value,
                             max_mutation_value=max_mutation_value,
                             **kwargs)


if __name__ == '__main__':
    from copy import copy

    pop_shape = (2, 4, 4)
    max_gene_value = 10
    min_gene_value = 0
    population_fraction = 1
    mutations_probability = (0.5, 0.5)
    min_mutation_value = 1
    max_mutation_value = 1

    population = B.randint(0, max_gene_value, pop_shape)
    RM = RandomMutations2D(population_fraction=population_fraction,
                           mutations_probability=mutations_probability,
                           min_gene_value=min_gene_value,
                           max_gene_value=max_gene_value,
                           min_mutation_value=min_mutation_value,
                           max_mutation_value=max_mutation_value,
                           debug=True)
    RM(population)

    chromosomes_sav = copy(population)
    cprint('[Initial genepool]', 'blue')
    cprint(chromosomes_sav, 'blue')
    population = RM(population)

    cprint('\n[Mutated genepool]', 'yellow')
    cprint(population, 'yellow')

    cprint('\n[Diff]', 'magenta')
    diff = population - chromosomes_sav
    cprint(diff, 'magenta')

    assert B.max(diff) <= max_mutation_value



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.engine import OP
from evoflow import backend as B


class DualCrossover(OP):
    def __init__(self, population_fraction, crossover_probability, **kwargs):
        """Perform Dual crossovers on a given population.

        Args:
            population_fraction (float): How many chromosomes
            should have a cross-over.

            crossover_probability (list(float)): What fraction of the
            gene should be affected by crossovers.

            debug (bool, optional): print debug information and function
            returns additional data.

        Returns:
            tensor: population with a crossover.

        See:
            https://en.wikipedia.org/wiki/Crossover_(genetic_algorithm)
        """

        if not (0 < population_fraction <= 1.0):
            raise ValueError("population_fraction must be in ]0. 1]")

        for val in crossover_probability:
            if not (0 < val <= 1.0):
                raise ValueError(
                    "crossover_size_fraction values must be between ]0. 1]")

        self.population_fraction = population_fraction
        self.crossover_probability = crossover_probability
        self.has_cache = False
        self.mutation_shape = []
        self.slices = None
        super(DualCrossover, self).__init__(**kwargs)

    def call(self, populations):
        results = []
        for population in populations:
            results.append(self.compute(population))
        return results

    def compute(self, population):

        # mix genomes
        population_copy = B.copy(population)
        population_copy = B.shuffle(population_copy)

        # how many chromosomes to crossover
        num_crossover_chromosomes = int(population.shape[0] *
                                        self.population_fraction)

        self.print_debug('num chromosomes', num_crossover_chromosomes)

        if not self.has_cache:
            self.print_debug("Caching fancy indexing")
            self.has_cache = True

            # compute the shape needed for the mutation
            mutations_shape = [num_crossover_chromosomes]
            for idx, frac in enumerate(self.crossover_probability):
                num_genes = int(population.shape[idx + 1] * frac)
                mutations_shape.append(num_genes)
            self.mutations_shape = mutations_shape
            self.print_debug("population_shape:", population.shape)
            self.print_debug("mutation_shape:", self.mutations_shape)

        else:
            self.print_debug("Using cached fancy indexing")

        # compute the fancy indexing dynamlically
        # ! the start point must be randomized
        slices = [slice(0, num_crossover_chromosomes)]
        for idx, crossover_size in enumerate(self.mutations_shape[1:]):
            # ! making indexing explicit as its a huge pitfall
            mutation_dim = idx + 1
            max_start = population.shape[mutation_dim] - crossover_size + 1
            start = B.randint(0, max_start, 1)[0]
            slices.append(slice(start, crossover_size + start))

        slices = tuple(slices)
        self.print_debug(slices)

        # crossover
        cross_section = population_copy[slices]
        population = B.assign(population, cross_section, slices)

        return population


class DualCrossover1D(DualCrossover):
    def __init__(self,
                 population_fraction=0.9,
                 crossover_probability=0.2,
                 **kwargs):
        if not isinstance(crossover_probability, float):
            raise ValueError('crossover_probability must be a float')

        super(DualCrossover1D,
              self).__init__(population_fraction=population_fraction,
                             crossover_probability=[crossover_probability],
                             **kwargs)


class DualCrossover2D(DualCrossover):
    def __init__(self,
                 population_fraction=0.9,
                 crossover_probability=(0.2, 0.2),
                 **kwargs):

        if len(crossover_probability) != 2:
            raise ValueError('crossover_probability must be of form (x, y)')
        super(DualCrossover2D,
              self).__init__(population_fraction=population_fraction,
                             crossover_probability=crossover_probability,
                             **kwargs)


class DualCrossover3D(DualCrossover):
    def __init__(self,
                 population_fraction=0.9,
                 crossover_probability=(0.2, 0.2, 0.2),
                 **kwargs):

        if len(crossover_probability) != 3:
            raise ValueError('crossover_probability must be of form (x, y, z)')

        super(DualCrossover3D,
              self).__init__(population_fraction=population_fraction,
                             crossover_probability=crossover_probability,
                             **kwargs)


if __name__ == '__main__':
    from copy import copy
    GENOME_SHAPE = (6, 4, 4)
    population = B.randint(0, 256, GENOME_SHAPE)
    population_fraction = 0.5
    max_crossover_size_fraction = (0.5, 0.5)
    print(population.shape)
    original_population = copy(population)
    population = DualCrossover2D(population_fraction,
                                 max_crossover_size_fraction,
                                 debug=True)(population)

    # diff matrix
    diff = B.clip(abs(population - original_population), 0, 1)
    print(diff)



from evoflow.engine import EvoFlow
from evoflow.ops import Input, RandomMutations2D, UniformCrossover2D
from evoflow.selection import SelectFittest
from evoflow.fitness import Sum
import evoflow.backend as B


def test_direct_2d():
    NUM_EVOLUTIONS = 2
    POPULATION_SIZE = 32
    GENE_SIZE = 10
    MAX_VAL = 256
    SHAPE = (POPULATION_SIZE, GENE_SIZE, GENE_SIZE)

    population = B.randint(0, MAX_VAL, SHAPE)
    inputs = Input(shape=SHAPE)
    x = RandomMutations2D(max_gene_value=1, min_gene_value=0)(inputs)
    outputs = UniformCrossover2D()(x)
    ef = EvoFlow(inputs, outputs)

    fitness_function = Sum(max_sum_value=GENE_SIZE)
    evolution_strategy = SelectFittest()

    ef.compile(evolution_strategy, fitness_function)
    ef.evolve(population, generations=NUM_EVOLUTIONS)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.ops import Input
from evoflow import backend as B


def test_call_vs_get():
    shape = (128, 64)
    population = B.randint(1, 10, shape=shape)
    inputs = Input(shape)
    inputs.assign(population)
    assert B.tensor_equal(inputs.get(), inputs.call(''))


def test_1d():
    shape = (128, 64)
    population = B.randint(1, 10, shape=shape)
    inputs = Input(shape)
    inputs.assign(population)
    assert B.tensor_equal(inputs.get(), population)


def test_2d():
    shape = (128, 64, 64)
    population = B.randint(1, 10, shape=shape)
    inputs = Input(shape)
    inputs.assign(population)
    assert B.tensor_equal(inputs.get(), population)


def test_non_tensor_input():
    shape = (2, 4)
    population = [[1, 2, 3, 4], [1, 2, 3, 4]]
    inputs = Input(shape)
    inputs.assign(population)
    res = inputs.get()
    assert B.is_tensor(res)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# - shared function --
# ! Dont import loader related funciton here - will break everything
# ! instead if needed (like set_backend()) put then in root __init__.py
from evoflow.config import floatx  # noqa: F401
from evoflow.config import set_floatx  # noqa: F401
from evoflow.config import intx  # noqa: F401
from evoflow.config import set_intx  # noqa: F401

# -initialization-
from .load_backend import tensor  # noqa: F401
from .load_backend import copy  # noqa: F401
from .load_backend import zeros  # noqa: F401
from .load_backend import ones  # noqa: F401
from .load_backend import fill  # noqa: F401
from .load_backend import normal  # noqa: F401
from .load_backend import range  # noqa: F401

# - reduce -
from .load_backend import prod  # noqa: F401
from .load_backend import max  # noqa: F401
from .load_backend import min  # noqa: F401
from .load_backend import sum  # noqa: F401
from .load_backend import mean  # noqa: F401
from .load_backend import sqrt  # noqa: F401

# - Manipulation -
from .load_backend import reverse  # noqa: F401
from .load_backend import roll  # noqa: F401
from .load_backend import assign  # noqa: F401
from .load_backend import tile  # noqa: F401
from .load_backend import concatenate  # noqa: F401

# - Utils -
from .load_backend import transpose  # noqa: F401
from .load_backend import cast  # noqa: F401
from .load_backend import dtype  # noqa: F401
from .load_backend import flatten  # noqa: F401
from .load_backend import as_numpy_array  # noqa: F401
from .load_backend import reshape  # noqa: F401
from .load_backend import is_tensor  # noqa: F401
from .load_backend import tensor_equal  # noqa: F401
from .load_backend import assert_near  # noqa: F401

# - Math -
from .load_backend import dot  # noqa: F401
from .load_backend import add  # noqa: F401
from .load_backend import subtract  # noqa: F401
from .load_backend import multiply  # noqa: F401
from .load_backend import divide  # noqa: F401
from .load_backend import mod  # noqa: F401
from .load_backend import clip  # noqa: F401
from .load_backend import abs  # noqa: F401
from .load_backend import broadcasted_norm  # noqa: F401
from .load_backend import norm  # noqa: F401

# - Randomness -
from .load_backend import randint  # noqa: F401
from .load_backend import shuffle  # noqa: F401
from .load_backend import full_shuffle  # noqa: F401

# - Indexing -
from .load_backend import take  # noqa: F401
from .load_backend import bottom_k_indices  # noqa: F401
from .load_backend import top_k_indices  # noqa: F401
from .load_backend import unique_with_counts  # noqa: F401

# - Statistical -
from .load_backend import bincount  # noqa: F401



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import os
# os.environ['evoflow_BACKEND'] = 'numpy'

from copy import copy
from termcolor import cprint
from evoflow import backend as B  # noqa: F402
from evoflow.ops import SingleCrossover1D, SingleCrossover2D
from evoflow.ops import SingleCrossover3D


def test_ND():
    "test various tensor size random"

    TEST_INPUTS = [
        [SingleCrossover1D, (2, 4), 0.5],
        [SingleCrossover2D, (2, 4, 4), (0.5, 0.5)],
        [SingleCrossover3D, (2, 4, 4, 4), (0.5, 0.5, 0.5)],
    ]

    for inputs in TEST_INPUTS:
        OP = inputs[0]
        pop_shape = inputs[1]
        mutations_probability = inputs[2]
        population_fraction = 1
        population = B.randint(0, 1024, pop_shape)

        # eager
        RM = OP(population_fraction=population_fraction,
                mutations_probability=mutations_probability)

        population = RM(population)
        assert B.is_tensor(population)
        assert population.shape == pop_shape

        # graph
        RM = OP(population_fraction=population_fraction,
                mutations_probability=mutations_probability)

        population = RM._call_from_graph(population)
        assert B.is_tensor(population)
        assert population.shape == pop_shape


def test_1D_shape():
    POPULATION_SHAPE = (64, 16)
    population = B.randint(0, 1024, POPULATION_SHAPE)
    population_fraction = 0.5
    crossover_size_fraction = 0.2

    original_population = copy(population)
    population = SingleCrossover1D(population_fraction,
                                   crossover_size_fraction,
                                   debug=1)(population)

    cprint(population, 'cyan')
    cprint(original_population, 'yellow')

    assert population.shape == POPULATION_SHAPE
    # measuring mutation rate
    diff = B.clip(abs(population - original_population), 0, 1)
    cprint('diff', 'cyan')
    cprint(diff, 'cyan')

    # row test
    num_ones_in_row = 0
    for col in diff:
        num_ones = list(col).count(1)
        num_ones_in_row = max(num_ones, num_ones_in_row)

    max_one_in_row = POPULATION_SHAPE[1] * crossover_size_fraction
    assert num_ones_in_row <= max_one_in_row
    assert num_ones_in_row

    # col
    diff = B.transpose(diff)
    num_ones_in_col = 0
    for col in diff:
        num_ones_in_col = max(list(col).count(1), num_ones_in_col)

    max_one_in_col = POPULATION_SHAPE[0] * population_fraction
    assert max_one_in_col - 2 <= num_ones_in_col <= max_one_in_col



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import evoflow.backend as B
from copy import copy
from evoflow.ops import UniformCrossover1D, UniformCrossover2D
from evoflow.ops import UniformCrossover3D


def test_uniform_2Dcrossover_randomness_shape():
    GENOME_SHAPE = (10, 4, 4)
    population = B.randint(0, 1024, GENOME_SHAPE)
    population_fraction = 0.5
    crossover_probability = (0.5, 0.5)

    original_population = copy(population)
    OP = UniformCrossover2D(population_fraction, crossover_probability)
    population = OP(population)

    diff = B.clip(abs(population - original_population), 0, 1)
    print(diff)
    expected_mutations = original_population.shape[0] * population_fraction

    mutated_chromosomes = []
    for c in diff:
        if B.max(c):
            mutated_chromosomes.append(c)
    num_mutations = len(mutated_chromosomes)

    # sometime we have a collision so we use a delta
    assert abs(num_mutations - expected_mutations) < 2

    mutated_rows = crossover_probability[0] * GENOME_SHAPE[1]
    mutated_cells = crossover_probability[0] * GENOME_SHAPE[2]
    for cidx, c in enumerate(mutated_chromosomes):
        mr = 0.0
        mc = 0.0
        for r in c:
            s = B.cast(B.sum(r), B.floatx())
            if s:
                mr += 1
                mc += s

        assert abs(mutated_rows - mr) < 2
        assert abs(B.cast(mutated_cells, B.floatx()) -
                   (mc / mutated_rows)) < 2.0  # noqa


def test_uniformcrossover2d_distribution():
    """check that every gene of the tensor are going to be flipped equally
    Note:
    # ! We need enough iterations and chromosomes to reduce collision
    # ! and ensure numerical stability
    """
    NUM_ITERATIONS = 1000
    GENOME_SHAPE = (20, 4, 4)
    population = B.randint(0, 1024, GENOME_SHAPE)
    population_fraction = 1
    crossover_probability = (0.5, 0.5)

    # each gene proba of being mutated 0.5*0.5 > 0.25
    # each chromosome proba of being mutated 1
    # => gene average hit rate: 1000 / (1/4)  ~250
    MIN_DIFF_BOUND = 180
    MAX_DIFF_BOUND = 320

    OP = UniformCrossover2D(population_fraction, crossover_probability)

    # diff matrix
    previous_population = copy(population)
    population = OP(population)
    diff = B.clip(abs(population - previous_population), 0, 1)
    for _ in range(NUM_ITERATIONS - 1):
        previous_population = copy(population)
        population = OP(population)

        curr_diff = B.clip(abs(population - previous_population), 0, 1)
        # acumulating diff matrix
        diff += curr_diff

    print(curr_diff)

    for c in diff:
        print(c)
        print('mean', B.mean(c), 'min', B.min(c), 'max', B.max(c))
        assert B.min(c) > MIN_DIFF_BOUND
        assert B.max(c) < MAX_DIFF_BOUND
        assert MIN_DIFF_BOUND < B.mean(c) < MAX_DIFF_BOUND


def test_ND():
    "test various tensor size random"

    TEST_INPUTS = [
        [UniformCrossover1D, (2, 4), 0.5],
        [UniformCrossover2D, (2, 4, 4), (0.5, 0.5)],
        [UniformCrossover3D, (2, 4, 4, 4), (0.5, 0.5, 0.5)],
    ]

    for inputs in TEST_INPUTS:
        OP = inputs[0]
        pop_shape = inputs[1]
        mutations_probability = inputs[2]
        population_fraction = 1
        population = B.randint(0, 1024, pop_shape)

        # eager
        RM = OP(population_fraction=population_fraction,
                mutations_probability=mutations_probability)

        population = RM(population)
        assert B.is_tensor(population)
        assert population.shape == pop_shape

        # graph
        RM = OP(population_fraction=population_fraction,
                mutations_probability=mutations_probability)

        population = RM._call_from_graph(population)
        assert B.is_tensor(population)
        assert population.shape == pop_shape



import evoflow.backend as B


def randint_population(shape, max_value, min_value=0):
    """Generate a random  population made of Integers

    Args:
        (set of ints): shape of the population. Its of the form
        (num_chromosomes, chromosome_dim_1, .... chromesome_dim_n)

        max_value (int): Maximum value taken by a given gene.

        min_value (int, optional): Min value a gene can take. Defaults to 0.

    Returns:
        Tensor: random population.
    """
    high = max_value + 1
    return B.randint(low=min_value, high=high, shape=shape)


def uniform_population(shape, dtype=B.intx()):
    """Generate a uniform population made of Integers. Uniform means that
    each chromosome contains only one time each value and each chromosome
    have them in different order.

    Args:
        (set of ints): shape of the population. Its of the form
        (num_chromosomes, chromosome size)

        dtype (str): tensor type

    Returns:
        Tensor: uniform population.
    """
    population = []
    chromosome = B.range(shape[1], dtype=dtype)
    for i in range(shape[0]):
        chromosome = B.shuffle(chromosome)
        population.append(chromosome)
    return B.tensor(population)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from evoflow.io import print_debug
import evoflow.backend as B
from collections import defaultdict


class FitnessFunction(object):
    """Fitness base classe.
    To be implemented by subclasses:
    * `call()`: Contains the logic for fitness calculation using `genes.
    """
    def __init__(self, **kwargs):
        self.debug = kwargs.get('debug', False)
        self._metrics = defaultdict(dict)

    def __call__(self, population):
        return self.call(population)

    @abc.abstractmethod
    def call(self, population):
        """Invokes the `Fitness` instance.

        Args:
            population: Tensor containing the population to assess
        """
        NotImplementedError('Must be implemented in subclasses.')

    def _flatten_population(self, population):
        """Convert the population to a 1D array as many ops (e.g norm) don't
        work on Ndimension
        """
        if len(population.shape) < 3:
            return population
        num_chromosomes = population.shape[0]
        flattened_size = int(B.prod(B.tensor(population.shape[1:])))
        return B.reshape(population, (num_chromosomes, flattened_size))

    def record_metric(self, name, value, group='default'):
        """Record metrics

        Args:
            name (str): name of the metrics.
            value (float): value to be recorded.
            group (str): group the metric belongs to. Used to group metrics
            together at display time.
        """
        self._metrics[group][name] = float(value)

    def get_metrics(self):
        """Return the lastest value of the recorded metrics
        """
        return self._metrics

    def print_debug(self, *msg):
        "output debug message"
        if self.debug:
            name = self.__class__.__name__
            print_debug(name, msg)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from termcolor import cprint


def arr2d():
    return [[1, 2, 3], [4, 5, 6], [7, 8, 9]]


def arr2dlarge():
    return [[1, 2, 3, 10, 11, 12], [4, 5, 6, 13, 14, 15],
            [7, 8, 9, 16, 17, 18]]


def test_shuffle_axis0(backends):
    for B in backends:
        t = B.tensor(arr2d())

        # give it 20 tries to ensure consistency
        for _ in range(20):
            t = B.shuffle(t)
            if t[0][0] != 1 or t[1][0] != 4:
                break

        assert t[0][0] in [1, 4, 7]
        assert t[1][0] in [1, 4, 7]
        assert t[2][0] in [1, 4, 7]

        assert t[0][0] != 1 or t[1][0] != 4


def test_shuffle_axis1(backends):
    for B in backends:
        t = B.tensor(arr2d())
        cprint(t, 'blue')

        # give it 20 tries to ensure consistency
        for _ in range(20):
            t = B.shuffle(t, axis=1)
            if t[0][0] != 1 or t[1][0] != 4:
                break

        cprint(t, 'green')

        assert t[0][0] in [1, 2, 3]
        assert t[1][0] in [4, 5, 6]
        assert t[2][0] in [7, 8, 9]

        assert t[0][0] != 1 or t[1][0] != 4


def test_full_shuffle(backends):
    for B in backends:
        t = B.tensor(arr2d())
        cprint(t, 'blue')

        # give it multiple try as identity is a valid shuffle

        for _ in range(20):
            t = B.full_shuffle(t)
            if (t[0][0] != 1 or t[1][0] != 4) and (t[0][1] != 2
                                                   or t[1][1] != 5):  # noqa

                break

        cprint(t, 'green')
        assert (t[0][0] != 1 or t[1][0] != 4)
        assert (t[0][1] != 2 or t[1][1] != 5)


def test_randint_1D(backends):
    for B in backends:
        t = B.randint(0, 11, shape=10)
        assert t.shape == (10, )
        assert B.max(t) <= 10
        assert B.min(t) >= 0


def test_single_number(backends):
    for B in backends:
        t = B.randint(11)
        assert t <= 10
        assert t >= 0

        t = B.randint(5, 11)
        assert t <= 10
        assert t >= 5


def test_randint_2SD(backends):
    for B in backends:
        t = B.randint(0, 11, shape=(10, 20))
        assert t.shape == (10, 20)
        assert B.max(t) <= 10
        assert B.min(t) >= 0



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from termcolor import cprint
from evoflow.engine import OP
from evoflow import backend as B


class UniformCrossover(OP):
    def __init__(self, population_fraction, crossover_probability, **kwargs):
        """Perform uniform crossovers on a given population.

        Args:
            population_fraction (float): How many chromosomes
            should have a cross-over.

            crossover_probability (list(float)): What fraction of the
            gene should be affected by crossovers.

            debug (bool, optional): print debug information and function
            returns additional data.

        Returns:
            tensor: population with a crossover.

        See:
            https://en.wikipedia.org/wiki/Crossover_(genetic_algorithm)
        """

        if not (0 < population_fraction <= 1.0):
            raise ValueError("population_fraction must be in ]0. 1]")

        for val in crossover_probability:
            if not (0 < val <= 1.0):
                raise ValueError(
                    "values in crossover_probability must be between ]0. 1]")

        self.population_fraction = population_fraction
        self.crossover_probability = crossover_probability
        self.x_matrix = None
        self.has_cache = False
        self.mutation_shape = []
        super(UniformCrossover, self).__init__(**kwargs)

    def call(self, populations):
        results = []
        for population in populations:
            results.append(self.compute(population))
        return results

    def compute(self, population):

        # mix genomes
        population_copy = B.copy(population)
        population_copy = B.shuffle(population_copy)

        # how many chromosomes to crossover?
        num_crossovers = int(population.shape[0] * self.population_fraction)
        self.print_debug("population size %s" % population.shape[0])
        self.print_debug("num_crossovers %s" % num_crossovers)

        # crossover matrix
        if not self.has_cache:
            self.print_debug("Creating x_matrix and mutation_matrix")
            self.has_cache = True
            self.x_matrix = B.zeros(population.shape, dtype=B.intx())

            # We need to accounting for the fact that the population
            # can be of rank N which makes the fancy indexing painful.
            # we need a shape for the mutation which is this:
            # [num_crossover, num_mutation, ..., num_mutations]
            mutations_shape = [num_crossovers]
            for idx, frac in enumerate(self.crossover_probability):
                num_genes = int(self.x_matrix.shape[idx + 1] * frac)
                mutations_shape.append(num_genes)
            self.print_debug("mutation_shape: %s" % mutations_shape)
            self.mutation_shape = mutations_shape

            # create tensor
            mutations = B.ones(mutations_shape)

            # compute the fancy indexing dynamically
            slices = []
            for size in mutations_shape:
                slices.append(slice(0, size))
            slices = tuple(slices)

            # injecting mutations
            self.x_matrix = B.assign(self.x_matrix, mutations, slices)
        else:
            self.print_debug("Using cached matrix")

        self.x_matrix = B.full_shuffle(self.x_matrix)

        # invert crossover matrix
        inv_x_matrix = B.abs((self.x_matrix) - 1)

        # copy chromosomes that stays the same
        population = population * inv_x_matrix

        # add the mutations
        population += (population_copy * self.x_matrix)

        return population


class UniformCrossover1D(UniformCrossover):
    def __init__(self,
                 population_fraction=0.9,
                 crossover_probability=0.2,
                 **kwargs):
        if not isinstance(crossover_probability, float):
            raise ValueError('crossover_probability must be a float')

        super(UniformCrossover1D,
              self).__init__(population_fraction=population_fraction,
                             crossover_probability=[crossover_probability],
                             **kwargs)


class UniformCrossover2D(UniformCrossover):
    def __init__(self,
                 population_fraction=0.9,
                 crossover_probability=(0.2, 0.2),
                 **kwargs):

        if len(crossover_probability) != 2:
            raise ValueError('crossover_size_fraction must be of form (x, y)')
        super(UniformCrossover2D,
              self).__init__(population_fraction=population_fraction,
                             crossover_probability=crossover_probability,
                             **kwargs)


class UniformCrossover3D(UniformCrossover):
    def __init__(self,
                 population_fraction=0.9,
                 crossover_probability=(0.2, 0.2, 0.2),
                 **kwargs):
        if len(crossover_probability) != 3:
            raise ValueError(
                'crossover_probability must be of form (x, y, z)')  # noqa
        super(UniformCrossover3D,
              self).__init__(population_fraction=population_fraction,
                             crossover_probability=crossover_probability,
                             **kwargs)


if __name__ == '__main__':
    from copy import copy
    GENOME_SHAPE = (10, 4, 4)
    population = B.randint(0, 1024, GENOME_SHAPE)
    population_fraction = 0.5
    crossover_probability = (0.5, 0.5)

    print(population.shape)
    # peforming crossover
    original_population = copy(population)
    population = UniformCrossover2D(population_fraction,
                                    crossover_probability)(population)

    # diff matrix
    diff = B.clip(abs(population - original_population), 0, 1)
    print(diff)
    # expected mutated chromosomes
    expected_mutated = population.shape[0] * population_fraction
    cprint(
        "Expected mutated chromosome:%d (+/- 1 due to collision)" %
        (expected_mutated), 'cyan')
    # select mutated chromosomes
    mutated_chromosomes = []
    for c in diff:
        if B.max(c):
            mutated_chromosomes.append(c)
    cprint("mutated chromosome:%d" % (len(mutated_chromosomes)), 'magenta')

    cprint("[example of mutated chromosome]", 'yellow')
    cprint(mutated_chromosomes[0], 'cyan')

    OP = UniformCrossover2D(population_fraction, crossover_probability)
    res = OP(population)
    for _ in range(100):
        prev = copy(res)
        res = OP(population)
        # acumulating diff matrix
        diff += B.clip(abs(res - prev), 0, 1)

    cprint("[cumulative diff matrix 100 runs]", 'yellow')
    cprint(diff[0], 'cyan')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import time
from setuptools import setup
from setuptools import find_packages

long_description = open("README.md").read()
version = '0.4.1_%s' % int(time())

setup(name='evoflow',
      version=version,
      description='Genetic algorithms for humans',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Elie Bursztein',
      author_email='evoflow@google.com',
      url='https://github.com/google/evoflow',
      license='Apache 2',
      python_requires='>=3',
      package_data={"": ["*.json"]},
      install_requires=[
          'networkx', 'numpy', 'tqdm', 'tabulate', 'termcolor', 'matplotlib',
          'altair', 'tensorflow>=2.2.0', 'pandas'
      ],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers', 'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      packages=find_packages())



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_prod(backends):
    inputs = [1, 2, 3, 4]
    for B in backends:
        tensor = B.tensor(inputs)
        assert B.prod(tensor) == 24


def test_sum(backends):
    inputs = [1, 2, 3, 4]
    for B in backends:
        tensor = B.tensor(inputs)
        assert B.sum(tensor) == 10


def test_max(backends):
    inputs = [1, 2, 3, 4]
    for B in backends:
        tensor = B.tensor(inputs)
        assert B.max(tensor) == 4


def test_min(backends):
    inputs = [1, 2, 3]
    for B in backends:
        tensor = B.tensor(inputs)
        assert B.min(tensor) == 1


def test_mean(backends):
    inputs = [1, 2, 3]
    for B in backends:
        tensor = B.tensor(inputs)
        assert B.mean(tensor) == 2



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.io import print_debug


class Callback(object):
    """Abstract base class used to build new callbacks.
 """
    def __init__(self):
        pass

    def on_generation_begin(self, generation):
        """Called at the start of a generation.

        Args:
            generation_idx (int): Generation.
        """
        pass

    def on_generation_end(self, generation, metrics, fitness_scores,
                          populations):
        """Called at the end of a generation.

        Args:
            population (Tensor): populations after the generation evolution.
        """
        pass

    def print_debug(self, *msg):
        "output debug message"
        if self.debug:
            name = self.__class__.__name__
            print_debug(name, msg)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import os
# os.environ['evoflow_BACKEND'] = 'numpy'
from termcolor import cprint
from copy import copy
from evoflow import backend as B
from evoflow.ops import DualCrossover1D, DualCrossover2D
from evoflow.ops import DualCrossover3D


def test_ND():
    "test various tensor size random"

    TEST_INPUTS = [
        [DualCrossover1D, (2, 4), 0.5],
        [DualCrossover2D, (2, 4, 4), (0.5, 0.5)],
        [DualCrossover3D, (2, 4, 4, 4), (0.5, 0.5, 0.5)],
    ]

    for inputs in TEST_INPUTS:
        OP = inputs[0]
        pop_shape = inputs[1]
        mutations_probability = inputs[2]
        population_fraction = 1
        population = B.randint(0, 1024, pop_shape)

        # eager
        RM = OP(population_fraction=population_fraction,
                mutations_probability=mutations_probability)

        population = RM(population)
        assert B.is_tensor(population)
        assert population.shape == pop_shape

        # graph
        RM = OP(population_fraction=population_fraction,
                mutations_probability=mutations_probability)

        population = RM._call_from_graph(population)
        assert B.is_tensor(population)
        assert population.shape == pop_shape


def test_crossover1D_output_shape():
    POPULATION_SHAPE = (8, 6)
    population = B.randint(0, 1024, POPULATION_SHAPE)
    population_fraction = 0.5
    mutations_probability = 0.2

    original_population = copy(population)
    population = DualCrossover1D(population_fraction,
                                 mutations_probability,
                                 debug=True)(population)

    cprint(population, 'cyan')
    cprint(original_population, 'yellow')

    assert population.shape == POPULATION_SHAPE
    # measuring mutation rate
    diff = B.clip(abs(population - original_population), 0, 1)

    # row test
    num_ones_in_row = 0
    for col in diff:
        num_ones_in_row = max(list(col).count(1), num_ones_in_row)

    max_one_in_row = int(POPULATION_SHAPE[1] * mutations_probability)
    assert num_ones_in_row == max_one_in_row
    assert num_ones_in_row


def test_dualcrossover2d_distribution():
    """check that every gene of the tensor are going to be flipped equally
    Note:
    # ! We need enough iterations and chromosomes to reduce collision
    # ! and ensure numerical stability
    """
    NUM_ITERATIONS = 1000
    GENOME_SHAPE = (100, 4, 4)
    population = B.randint(0, 1024, GENOME_SHAPE)
    population_fraction = 1
    crossover_probability = (0.5, 0.5)

    OP = DualCrossover2D(population_fraction, crossover_probability)

    # diff matrix
    previous_population = copy(population)
    population = OP(population)
    diff = B.clip(abs(population - previous_population), 0, 1)
    print(diff)

    for _ in range(NUM_ITERATIONS - 1):
        previous_population = copy(population)
        population = OP(population)

        curr_diff = B.clip(abs(population - previous_population), 0, 1)
        # acumulating diff matrix
        diff += curr_diff

    # print(curr_diff)

    for c in diff:
        print(c)
        print('mean', B.mean(c), 'min', B.min(c), 'max', B.max(c))
        assert B.min(c) > 50
        assert B.max(c) < NUM_ITERATIONS // 2
        assert 200 < B.mean(c) < NUM_ITERATIONS // 2



from evoflow.ops import Shuffle
from evoflow.population import randint_population
import evoflow.backend as B


def test_shuffle():

    SHAPES = [(2, 4), (2, 4, 4), (2, 4, 4, 4)]

    for shape in SHAPES:
        population = randint_population(shape, max_value=255)
        previous_population = B.copy(population)
        population = Shuffle(1, debug=True)(population)
        diff = B.clip(abs(population - previous_population), 0, 1)

        assert B.is_tensor(population)
        assert population.shape == shape
        assert B.sum(diff) > int(B.prod(shape) / 2)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.engine import EvoFlow
import evoflow.backend as B
from evoflow.ops import Input, RandomMutations1D, UniformCrossover1D
from evoflow.selection import SelectFittest
from evoflow.fitness import Sum
from evoflow.population import randint_population
from evoflow.callbacks import DummyCallback


def test_helloworld():
    "Solve the MAXONE problem"
    NUM_EVOLUTIONS = 10
    SHAPE = (512, 1024)
    MAX_VAL = 1

    population = randint_population(SHAPE, MAX_VAL)

    inputs = Input(SHAPE)
    x = RandomMutations1D(max_gene_value=1)(inputs)
    outputs = UniformCrossover1D()(x)
    gf = EvoFlow(inputs, outputs, debug=0)
    fitness_function = Sum(max_sum_value=SHAPE[1])
    evolution_strategy = SelectFittest()
    gf.compile(evolution_strategy, fitness_function)
    gf.summary()
    results = gf.evolve(population,
                        generations=NUM_EVOLUTIONS,
                        callbacks=[DummyCallback()])
    assert results

    metrics = results.get_metrics_history()

    # check metrics
    for metric_grp, data in metrics.items():

        for metric_name, vals in data.items():
            assert isinstance(vals, list)
            if metric_grp == 'Fitness function':
                assert len(vals) == 10
                assert isinstance(vals[9], float)

                assert 'mean' in data
                assert 'max' in data

    # assert the engine solved the (Trivial) problem
    max_fitness = metrics['Fitness function']['max']
    assert max(max_fitness) == 1
    assert max_fitness[9] == 1
    assert min(max_fitness) < 1  # max sure we did improve

    # check population value
    population = results.get_populations()
    assert (population.shape) == SHAPE
    assert B.tensor_equal(population[0], B.tensor([1] * SHAPE[1]))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.fitness import Sum
import evoflow.backend as B


def test_sum2d():
    t = [[[1, 1, 1], [1, 1, 1], [1, 1, 1]], [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
         [[1, 1, 1], [1, 1, 1], [1, 1, 1]]]
    inputs = B.tensor(t)
    print(inputs)
    result = Sum().call(inputs)
    assert result.shape == (3, )
    print(result)

    expected = B.tensor([9, 9, 9])
    assert B.tensor_equal(result, expected)


def test_max_gene_val_2d():
    MAX_VAL = 10
    t = B.randint(0, MAX_VAL + 1, (10, 10, 10))
    max_sum_value = MAX_VAL * 10 * 10
    v = Sum(max_sum_value=max_sum_value).call(t)
    assert v.shape == (10, )
    for t in v:
        assert t < 1



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import random
import hashlib
from time import time
from evoflow.utils import box, unbox
from evoflow.io import print_debug
from evoflow import backend as B


class OP(object):
    "Base class for all operations"

    def __init__(self, **kwargs):

        # naming
        self.op_type = self.__class__.__name__
        self.idx = kwargs.get('name', self._gen_name())
        self.debug = kwargs.get('debug', False)

        self.input_ops = []
        self.input_shapes = []  # track tensor size accross the ops.

    @abc.abstractmethod
    def call(self, chromosomes, **kwargs):
        """This is where the logic of the operation live.

        Args:
            chromosomes (ndarrays): Tensor or list of tensors
            **kwargs: Additional keyword arguments to be passed to `call()`.
        """
        return chromosomes

    def _gen_name(self):
        return self.op_type.lower() + "_" + self._gen_idx()

    def _call_from_graph(self, populations):
        "Function called during graph executions"
        populations = box(populations)
        return unbox(self.call(populations))

    def __call__(self, ops):

        # boxing inputs if needed
        ops = box(ops)

        # graph computation or eager?
        if self.debug:
            input_types = [type(op) for op in ops]
            self.print_debug('inputs type:%s' % (input_types))

        if issubclass(type(ops[0]), OP):
            # graph mode
            self.print_debug('graph mode')

            input_shapes = []
            for op in ops:
                assert issubclass(type(op), OP)
                self.input_ops.append(op)
                input_shapes.append(op.get_output_shapes())

            self.compute_output_shape(input_shapes)

            return self
        else:
            # eager mode
            self.print_debug('eager mode')

            # check inputs are valis
            for op in ops:
                if not B.is_tensor(op):
                    raise ValueError("Expecting list(tensors) or a tensor")

            # input shapes
            input_shapes = []
            for op in ops:
                input_shapes.append(op.shape)
            self.compute_output_shape(input_shapes)

            # compute concrete results
            results = self.call(ops)

            # unbox result if needed
            return unbox(results)

    def compute_output_shape(self, input_shapes):
        """Compute output shapes

        Args:
            list(tuples): input shapes

        Note:
            any kind of initialization depending of input shape can be done by
            redefining this function in the children class.
        """

        self.output_shapes = input_shapes

    def get_output_shapes(self):
        "return the shape of tensor returned by the op"
        return unbox(self.input_shapes)

    def _gen_idx(self):
        "generate a unique idx"
        idx = "%s%s" % (int(time()), random.randint(10000000, 90000000))
        idx = hashlib.md5(idx.encode()).hexdigest()[:6].upper()
        return idx

    def get_config(self):
        "export op config"
        config = {
            "node_type": self.op_type,
            "idx": self.idx,
            "inbound_ops": self.inbound_ops
        }
        return config

    def print_debug(self, *msg):
        "output debug message"
        if self.debug:
            print_debug(self.idx, msg)

    @classmethod
    def from_config(cls, config):
        "create an op from its config"
        return cls(**config)
